import dataclasses
import warnings
from logging import getLogger
from typing import Any, Dict, Iterable, List, Tuple

logger = getLogger(__name__)


class FlattenedAccess:
    """Allows flattened access to the attributes of all children dataclasses.

    This is meant to simplify the adoption of dataclasses for argument
    hierarchies, rather than a single-level dictionary.
    Dataclasses allow for easy, neatly separated arguments, but suffer from 2
    potential drawbacks:
    - When using a highly nested structure, having long accesses is annoying
    - The dictionary access syntax is often more natural than using getattr()
        when reading an attribute whose name is a variable.
    """

    def attributes(self, recursive: bool = True, prefix: str = "") -> Iterable[Tuple[str, Any]]:
        """Returns an Iterator over the attributes of the dataclass.

        [extended_summary]

        Parameters
        ----------
        - dataclass : Dataclass

            A dataclass type or instance.
        - recursive : bool, optional, by default True

            Whether or not to recurse and yield all the elements of the children
            dataclass attributes.
        - prefix : str, optional, by default ""

            A prefix to prepend to all the attribute names before yielding them.

        Returns
        -------
        Iterable[Tuple[str, Any]]
            An iterable of attribute names and values.

        Yields
        -------
        Iterable[Tuple[str, Any]]
            A Tuple of the form <Attribute name, attribute_value>.
        """
        for field in dataclasses.fields(self):
            if field.name not in self.__dict__:
                # the dataclass isn't yet instantiated, or the attr was deleted.
                continue
            # get the field value (without needless recursion)
            field_value = self.__dict__[field.name]

            yield prefix + field.name, field_value
            if recursive and dataclasses.is_dataclass(field_value):
                yield from FlattenedAccess.attributes(
                    field_value, recursive=True, prefix=prefix + field.name + "."
                )

    def __getattr__(self, name: str):
        """Retrieves the attribute on self, or recursively on the children.

        NOTE: `__getattribute__` is always called before `__getattr__`, hence we
        always get here because `self` does not have an attribute of `name`.
        """
        # potential parents and corresponding values.
        parents: List[str] = []
        values: List[Any] = []

        for attr_name, attr_value in FlattenedAccess.attributes(self):
            # if the attribute name's last part ends with `name`, we add it to
            # some list of potential parent attributes.
            name_parts = name.split(".")
            dest_parts = attr_name.split(".")
            if dest_parts[-len(name_parts) :] == name_parts:
                parents.append(attr_name)
                values.append(attr_value)

        if not parents:
            raise AttributeError(
                f"{type(self)} object has no attribute '{name}', "
                "and neither does any of its children attributes."
            )
        elif len(parents) > 1:
            raise AttributeError(
                f"Ambiguous Attribute access: name '{name}' may refer to:\n"
                + "\n".join(
                    f"- '{parent}' (with a value of: '{value}')"
                    for parent, value in zip(parents, values)
                )
            )
        else:
            return values[0]

    def __setattr__(self, name: str, value: Any):
        """Write the attribute in self or in the children that has it.

        If more than one child has attributes that match the given one, an
        `AttributeError` is raised.
        """
        # potential parents and corresponding values.
        parents: List[str] = []
        values: List[Any] = []

        field_names = {field.name for field in dataclasses.fields(self)}
        if name in field_names:
            object.__setattr__(self, name, value)
            return

        for attr_name, attr_value in self.attributes():
            # if the attribute name of the attribute ends with `name`, we add it
            # to some list of potential parent attributes.
            name_parts = name.split(".")
            dest_parts = attr_name.split(".")
            if dest_parts[-len(name_parts) :] == name_parts:
                parents.append(attr_name)
                values.append(attr_value)

        if not parents:
            # We set the value on the dataclass directly, since it wasn't found.
            warnings.warn(
                UserWarning(
                    f"Setting a new attribute '{name}' on the"
                    f" dataclass, but it does not have a field of the same name. \n"
                    f"(Consider adding a field '{name}' of type {type(value)} to "
                    f"{type(self)})"
                )
            )
            object.__setattr__(self, name, value)

        elif len(parents) > 1:
            # more than one parent (ambiguous).
            raise AttributeError(
                f"Ambiguous Attribute access: name '{name}' may refer to:\n"
                + "\n".join(
                    f"- '{parent}' (with a value of: '{value}')"
                    for parent, value in zip(parents, values)
                )
            )
        else:
            # We recursively set the attribute.
            attr_name = parents[0]
            lineage = attr_name.split(".")[:-1]
            parent: object = self
            for parent_name in lineage:
                # NOTE: we can't use getattr, otherwise we would recurse.
                parent = object.__getattribute__(parent, parent_name)
            # destination attribute name
            dest_name = name.split(".")[-1]
            # Set the attribute on the parent.
            object.__setattr__(parent, dest_name, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def asdict(self) -> Dict:
        return dataclasses.asdict(self)



import random


def set_seed(seed: int) -> None:
    random.seed(seed)
    try:
        import numpy as np

        np.random.seed(seed)
    except ImportError:
        pass

    try:
        import torch
    except ImportError:
        pass
    else:
        try:
            torch.manual_seed(seed)
            if torch.cuda.is_available():
                torch.cuda.manual_seed_all(seed)
        except AttributeError:
            pass



import argparse
import dataclasses
from argparse import Namespace
from dataclasses import dataclass
from pathlib import Path
from typing import Union

import pytest

import simple_parsing
from simple_parsing import ArgumentParser, choice
from simple_parsing.helpers import subparsers

from .testutils import TestSetup, raises, xfail


@dataclass
class TrainOptions:
    """Training Options"""

    lr: float = 1e-3
    train_path: Path = Path("./train")


@dataclass
class ValidOptions:
    """Validation Options"""

    test_path: Path = Path("./test")
    metric: str = "accuracy"


@dataclass
class GlobalOptions(TestSetup):
    """Global Options"""

    # mode, either Train or Valid.
    mode: Union[TrainOptions, ValidOptions] = subparsers(
        {
            "train": TrainOptions,
            "valid": ValidOptions,
        }
    )
    global_arg: str = "something"


# options = GlobalOptions.setup("--global_arg 123 --help")
# print(options)
# exit()


def test_required_subparser():
    options = GlobalOptions.setup("train --lr 0.1 --train_path ./bob")
    assert isinstance(options.mode, TrainOptions)
    assert options.mode.lr == 0.1
    assert options.mode.train_path == Path("./bob")

    options = GlobalOptions.setup("valid --metric some_metric --test_path ./john")
    assert isinstance(options.mode, ValidOptions)
    assert options.mode.metric == "some_metric"
    assert options.mode.test_path == Path("./john")


def test_help_text_works():
    from contextlib import suppress

    with suppress(SystemExit):
        GlobalOptions.setup("--help")
        GlobalOptions.setup("train --help")


@dataclass
class Start:
    """Start command"""

    value: str = "start command value"

    def execute(self, verbose=False):
        print(f"Start (verbose: {verbose})")
        return self.value


@dataclass
class Stop:
    """Stop command"""

    value: str = "stop command value"

    def execute(self, verbose=False):
        print(f"Stop (verbose: {verbose})")
        return self.value


@dataclass
class Push:
    """Example of a subcommand."""

    subcommand: Union[Start, Stop]
    some_flag_1: str = "command1"

    def execute(self, verbose=False):
        print(f"Push (verbose: {verbose})")
        return self.subcommand.execute(verbose=verbose)


@dataclass
class Pull:
    """Other Example of a subcommand."""

    subcommand: Union[Start, Stop]
    some_flag_2: str = "command2"

    def execute(self, verbose=False):
        print(f"Pull (verbose: {verbose})")
        return self.subcommand.execute(verbose=verbose)


@dataclass
class Program(TestSetup):
    """Some top-level command"""

    command: Union[Push, Pull]
    verbose: bool = False

    def execute(self):
        print(f"Program (verbose: {self.verbose})")
        return self.command.execute(verbose=self.verbose)


def test_command_tree():
    prog: Program = Program.setup("push start --value jack")
    assert isinstance(prog.command, Push)
    assert isinstance(prog.command.subcommand, Start)
    assert prog.command.subcommand.value == "jack"
    assert prog.execute() == "jack"


def test_experiments():
    from abc import ABC

    @dataclass
    class Experiment(ABC):
        dataset: str
        iid: bool = True

    @dataclass
    class Mnist(Experiment):
        dataset: str = "mnist"
        iid: bool = True

    @dataclass
    class MnistContinual(Experiment):
        dataset: str = "mnist"
        iid: bool = False

    @dataclass
    class Config:
        experiment: Experiment = subparsers(
            {
                "mnist": Mnist,
                "mnist_continual": MnistContinual,
            }
        )

    for field in dataclasses.fields(Config):
        assert simple_parsing.utils.is_subparser_field(field), field

    parser = ArgumentParser()
    parser.add_arguments(Config, "config")

    with raises(SystemExit):
        args = parser.parse_args()

    args = parser.parse_args("mnist".split())
    experiment = args.config.experiment
    assert isinstance(experiment, Mnist)
    assert experiment.dataset == "mnist"
    assert experiment.iid is True

    args = parser.parse_args("mnist_continual".split())
    experiment = args.config.experiment
    assert isinstance(experiment, MnistContinual)
    assert experiment.dataset == "mnist"
    assert experiment.iid is False


def test_subparser_rest_of_args_go_to_parent():
    @dataclass
    class Child:
        name: str = "Bob"
        age: int = 8

    @dataclass
    class Pet:
        kind: str = choice("cat", "dog", "fish")

    @dataclass
    class Parent(TestSetup):
        family: Union[Child, Pet]
        foo: bool = False
        income: float = 35_000.0

    p = Parent.setup(
        "pet --kind fish --foo=true --income 10_000",
        parse_known_args=True,
        attempt_to_reorder=True,
    )
    assert p == Parent(family=Pet(kind="fish"), foo=True, income=10_000.0)

    p = Parent.setup(
        "--income 10_000 pet --kind fish --foo=true",
        parse_known_args=True,
        attempt_to_reorder=True,
    )
    assert p == Parent(family=Pet(kind="fish"), foo=True, income=10_000.0)

    p = Parent.setup(
        "--income 10_000 --foo=true pet --kind fish",
        parse_known_args=True,
        attempt_to_reorder=True,
    )
    assert p == Parent(family=Pet(kind="fish"), foo=True, income=10_000.0)


@xfail(
    reason="TODO: Not sure how to fix this issue. Can only perform simple "
    "re-ordering for now. (as in the test above this one)"
)
def test_mixing_the_ordering():
    @dataclass
    class Child:
        name: str = "Bob"
        age: int = 8

    @dataclass
    class Pet:
        kind: str = choice("cat", "dog", "fish", default="cat")

    @dataclass
    class Parent(TestSetup):
        family: Union[Child, Pet] = subparsers(None, required=False)
        foo: bool = False
        income: float = 35_000.0

    p = Parent.setup(
        "--income 10_000 pet --foo --kind fish",
        parse_known_args=True,
        attempt_to_reorder=True,
    )
    assert p == Parent(family=Pet(kind="fish"), foo=True, income=10_000.0)


@xfail(reason="TODO")
def test_mixing_the_ordering_all_have_defaults():
    @dataclass
    class Child:
        name: str = "Bob"
        age: int = 8

    @dataclass
    class Pet:
        kind: str = choice("cat", "dog", "fish", default="cat")

    @dataclass
    class Parent(TestSetup):
        family: Union[Child, Pet] = subparsers(None, required=False)
        foo: bool = False
        income: float = 35_000.0

    p = Parent.setup(
        "--income 10_000 pet --foo --kind fish",
        parse_known_args=True,
        attempt_to_reorder=True,
    )
    assert p == Parent(family=Pet(kind="fish"), foo=True, income=10_000.0)


def test_argparse_version_giving_extra_args_to_parent():

    parser = argparse.ArgumentParser()
    parser.add_argument("--foo", type=int, default=3)

    subparsers = parser.add_subparsers(title="foo_command")

    subparser = subparsers.add_parser("boo")
    subparser.add_argument("--bar", type=int, default=1)
    subparser.add_argument("--baz", type=int, default=4)

    args = parser.parse_args("--foo 1 boo --bar 2 --baz 3".split())
    assert args == Namespace(foo=1, bar=2, baz=3)

    args = parser.parse_known_args("boo --bar 2 --baz 3 --foo 1".split())
    assert args == (Namespace(foo=3, bar=2, baz=3), ["--foo", "1"])


@pytest.mark.filterwarnings("ignore:Unparsed arguments")
def test_simpleparse_version_giving_extra_args_to_parent():
    parser = simple_parsing.ArgumentParser()
    parser.add_argument("--foo", type=int, default=3)
    assert not parser._subparsers
    subparsers = parser.add_subparsers(title="foo_command")

    subparser = subparsers.add_parser("boo")
    subparser.add_argument("--bar", type=int, default=1)
    subparser.add_argument("--baz", type=int, default=4)

    args = parser.parse_args("--foo 1 boo --bar 2 --baz 3".split())
    assert args == Namespace(foo=1, bar=2, baz=3)

    args = parser.parse_known_args("boo --bar 2 --baz 3 --foo 1".split(), attempt_to_reorder=True)
    assert args == (Namespace(foo=1, bar=2, baz=3), [])


if __name__ == "__main__":
    import sys

    print("ARGS:", " ".join(sys.argv[1:]))
    prog: Program = Program.setup(" ".join(sys.argv[1:]))
    print(prog)
    print(prog.execute())
    exit()



from __future__ import annotations

import os
import shlex
import string
import sys
from contextlib import contextmanager, redirect_stderr
from io import StringIO
from typing import Any, Callable, Generic, List, TypeVar, cast

import pytest

import simple_parsing
from simple_parsing import ConflictResolution, DashVariant, ParsingError
from simple_parsing.utils import camel_case
from simple_parsing.wrappers.field_wrapper import ArgumentGenerationMode, NestedMode

xfail = pytest.mark.xfail
parametrize = pytest.mark.parametrize


def xfail_param(*args, reason: str):
    if len(args) == 1 and isinstance(args[0], tuple):
        args = args[0]
    return pytest.param(*args, marks=pytest.mark.xfail(reason=reason))


Dataclass = TypeVar("Dataclass")


@contextmanager
def raises(exception: type[Exception] = ParsingError, match=None, code: int | None = None):
    with pytest.raises(exception, match=match):
        yield


@contextmanager
def exits_and_writes_to_stderr(match: str = ""):
    s = StringIO()
    with redirect_stderr(s), raises(SystemExit):
        yield
    s.seek(0)
    err_string = s.read()
    if match:
        assert match in err_string, err_string
    else:
        assert err_string, err_string


@contextmanager
def raises_missing_required_arg(args: str | list[str] = ""):
    args_str = " ".join(args) if isinstance(args, list) else args
    with exits_and_writes_to_stderr("the following arguments are required: " + args_str):
        yield


@contextmanager
def raises_invalid_choice():
    with exits_and_writes_to_stderr("invalid choice:"):
        yield


@contextmanager
def raises_expected_n_args(n: int | str):
    with exits_and_writes_to_stderr(f"expected {n} argument{'s' if n != 1 else ''}"):
        yield


@contextmanager
def raises_unrecognized_args(*args: str):
    with exits_and_writes_to_stderr("unrecognized arguments: " + " ".join(args or [])):
        yield


def assert_help_output_equals(actual: str, expected: str) -> None:
    # Replace the start with `prog`, since the test runner might not always be
    # `pytest`, could also be __main__ when debugging with VSCode
    prog = sys.argv[0].split("/")[-1]
    if prog != "pytest":
        expected = expected.replace("usage: pytest", f"usage: {prog}")
    remove = string.punctuation + string.whitespace

    if "optional arguments" in expected and sys.version_info[:2] >= (3, 10):
        expected = expected.replace("optional arguments", "options")

    actual_str = "".join(actual.split())
    actual_str = actual.translate(str.maketrans("", "", remove))
    expected_str = expected.translate(str.maketrans("", "", remove))
    assert actual_str == expected_str, "\n" + "\n".join([actual_str, expected_str])


T = TypeVar("T")


class TestParser(simple_parsing.ArgumentParser, Generic[T]):
    __test__ = False
    """ A parser subclass just used for testing.
    Makes the retrieval of the arguments a bit easier to read.
    """

    def __init__(self, *args, **kwargs):
        self._current_dest = None
        self._current_dataclass = None
        super().__init__(*args, **kwargs)

    def add_arguments(self, dataclass: type, dest, prefix="", default=None):
        if self._current_dest == dest and self._current_dataclass == dataclass:
            return  # already added arguments for that dataclass.
        self._current_dest = dest
        self._current_dataclass = dataclass
        return super().add_arguments(dataclass, dest, prefix=prefix, default=default)

    def __call__(self, args: str) -> T:
        namespace = self.parse_args(shlex.split(args))
        value = getattr(namespace, self._current_dest)
        value = cast(T, value)
        return value


def using_simple_api() -> bool:
    return os.environ.get("SIMPLE_PARSING_API", "simple") == "simple"


class TestSetup:
    @classmethod
    def setup(
        cls: type[Dataclass],
        arguments: str | None = "",
        dest: str | None = None,
        default: Dataclass | None = None,
        conflict_resolution_mode: ConflictResolution = ConflictResolution.AUTO,
        add_option_string_dash_variants: DashVariant = DashVariant.AUTO,
        parse_known_args: bool = False,
        attempt_to_reorder: bool = False,
        *,
        argument_generation_mode: ArgumentGenerationMode = ArgumentGenerationMode.FLAT,
        nested_mode: NestedMode = NestedMode.DEFAULT,
        **kwargs,
    ) -> Dataclass:
        """Basic setup for a test.

        Keyword Arguments:
            arguments {Optional[str]} -- The arguments to pass to the parser (default: {""})
            dest {Optional[str]} -- the attribute where the argument should be stored. (default: {None})

        Returns:
            {cls}} -- the class's type.
        """

        # TODO: Add a switch here, that uses either the 'simpler' API: `simple_parsing.parse` or
        # this more verbose API: `simple_parsing.ArgumentParser().parse_args()`
        dest = dest or camel_case(cls.__name__)

        if using_simple_api():
            common_kwargs = dict(
                config_class=cls,
                config_path=None,
                args=arguments,
                default=default,
                add_config_path_arg=None,
                nested_mode=nested_mode,
                dest=dest,
                add_option_string_dash_variants=add_option_string_dash_variants,
                conflict_resolution=conflict_resolution_mode,
                argument_generation_mode=argument_generation_mode,
            )
            if parse_known_args:
                instance, unknown_args = simple_parsing.parse_known_args(
                    **common_kwargs,
                    attempt_to_reorder=attempt_to_reorder,
                    **kwargs,
                )
            else:
                instance = simple_parsing.parse(**common_kwargs, **kwargs)
        else:
            parser = simple_parsing.ArgumentParser(
                conflict_resolution=conflict_resolution_mode,
                add_option_string_dash_variants=add_option_string_dash_variants,
                argument_generation_mode=argument_generation_mode,
                nested_mode=nested_mode,
                **kwargs,
            )
            if dest is None:
                dest = camel_case(cls.__name__)

            parser.add_arguments(cls, dest=dest, default=default)

            if arguments is None:
                if parse_known_args:
                    args = parser.parse_known_args(attempt_to_reorder=attempt_to_reorder)
                else:
                    args = parser.parse_args()
            else:
                splits = shlex.split(arguments)
                if parse_known_args:
                    args, unknown_args = parser.parse_known_args(
                        splits, attempt_to_reorder=attempt_to_reorder
                    )
                else:
                    args = parser.parse_args(splits)
            assert hasattr(args, dest), f"attribute '{dest}' not found in args {args}"
            instance: Dataclass = getattr(args, dest)  # type: ignore
            delattr(args, dest)

            # If there are subgroups, we can allow an extra "subgroups" attribute, otherwise we don't
            # expect any other arguments.
            args_dict = vars(args).copy()
            args_dict.pop("subgroups", None)

            assert (
                not args_dict
            ), f"Namespace has leftover garbage values (besides subgroups): {args}"

        instance = cast(Dataclass, instance)
        return instance

    @classmethod
    def setup_multiple(
        cls: type[Dataclass], num_to_parse: int, arguments: str | None = ""
    ) -> tuple[Dataclass, ...]:
        conflict_resolution_mode: ConflictResolution = ConflictResolution.ALWAYS_MERGE

        parser = simple_parsing.ArgumentParser(conflict_resolution=conflict_resolution_mode)
        class_name = camel_case(cls.__name__)
        for i in range(num_to_parse):
            parser.add_arguments(cls, f"{class_name}_{i}")

        if arguments is None:
            args = parser.parse_args()
        else:
            splits = shlex.split(arguments)
            args = parser.parse_args(splits)

        return tuple(getattr(args, f"{class_name}_{i}") for i in range(num_to_parse))

    @classmethod
    def get_help_text(
        cls,
        argv: str | None = None,
        multiple=False,
        conflict_resolution_mode: ConflictResolution = ConflictResolution.AUTO,
        add_option_string_dash_variants=DashVariant.AUTO,
        **parser_kwargs,
    ) -> str:
        import contextlib
        from io import StringIO

        f = StringIO()

        if argv is None:
            argv = "--help"
        elif not argv.endswith("--help"):
            argv = argv + " --help"

        with contextlib.suppress(SystemExit), contextlib.redirect_stdout(f):
            _ = cls.setup(
                argv,
                conflict_resolution_mode=conflict_resolution_mode,
                add_option_string_dash_variants=add_option_string_dash_variants,
                **parser_kwargs,
            )
        s = f.getvalue()
        return s


ListFormattingFunction = Callable[[List[Any]], str]
ListOfListsFormattingFunction = Callable[[List[List[Any]]], str]


def format_list_using_spaces(value_list: list[Any]) -> str:
    return " ".join(str(p) for p in value_list)


def format_list_using_brackets(value_list: list[Any]) -> str:
    return f"[{','.join(str(p) for p in value_list)}]"


def format_list_using_single_quotes(value_list: list[Any]) -> str:
    return f"'{format_list_using_spaces(value_list)}'"


def format_list_using_double_quotes(value_list: list[Any]) -> str:
    return f'"{format_list_using_spaces(value_list)}"'


def format_lists_using_brackets(list_of_lists: list[list[Any]]) -> str:
    return " ".join(format_list_using_brackets(value_list) for value_list in list_of_lists)


def format_lists_using_double_quotes(list_of_lists: list[list[Any]]) -> str:
    return " ".join(format_list_using_double_quotes(value_list) for value_list in list_of_lists)


def format_lists_using_single_quotes(list_of_lists: list[list[Any]]) -> str:
    return " ".join(format_list_using_single_quotes(value_list) for value_list in list_of_lists)



import argparse
import shlex
from dataclasses import dataclass
from test import raises_missing_required_arg

from simple_parsing import ArgumentParser
from simple_parsing.utils import str2bool


def test_suppress_default(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type
        b: int = 5

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default=argparse.SUPPRESS)

    parsed_args, unparsed_args = parser.parse_known_args(shlex.split(f"--a {passed_value} --b 0"))
    assert parsed_args == argparse.Namespace(some_class={"a": expected_value, "b": 0})
    assert unparsed_args == []

    parsed_args, unparsed_args = parser.parse_known_args(shlex.split(f"--a {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": expected_value})
    assert unparsed_args == []

    assert parser.parse_args("") == argparse.Namespace()
    assert parser.parse_args(shlex.split("--b 0")) == argparse.Namespace(some_class={"b": 0})


def test_with_unsuppressed_dataclass(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass1:
        a: some_type
        b: int = 5

    @dataclass
    class SomeClass2:
        c: some_type
        d: int = 5

    parser = ArgumentParser()
    parser.add_arguments(SomeClass1, dest="some_class1", default=argparse.SUPPRESS)
    parser.add_arguments(SomeClass2, dest="some_class2")

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value} --b 0 --c {passed_value}"))
    assert parsed_args == argparse.Namespace(
        some_class1={"a": expected_value, "b": 0},
        some_class2=SomeClass2(c=expected_value),
    )

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value} --c {passed_value}"))
    assert parsed_args == argparse.Namespace(
        some_class1={"a": expected_value},
        some_class2=SomeClass2(c=expected_value),
    )

    with raises_missing_required_arg("-c/--c"):
        parser.parse_args(shlex.split(f"--a {passed_value}"))

    with raises_missing_required_arg("-c/--c"):
        parser.parse_args("")

    assert parser.parse_args(shlex.split(f"--c {passed_value}")) == argparse.Namespace(
        some_class2=SomeClass2(expected_value)
    )


def test_conflict_with_unsuppressed_dataclass(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type
        b: int = 5

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class1", default=argparse.SUPPRESS)
    parser.add_arguments(SomeClass, dest="some_class2")

    parsed_args = parser.parse_args(
        shlex.split(
            f"--some_class1.a {passed_value} --some_class1.b 0 --some_class2.a {passed_value}"
        )
    )
    assert parsed_args == argparse.Namespace(
        some_class1={"a": expected_value, "b": 0},
        some_class2=SomeClass(expected_value),
    )

    parsed_args, unparsed_args = parser.parse_known_args(
        shlex.split(f"--a {passed_value} --some_class2.a {passed_value}")
    )
    assert parsed_args == argparse.Namespace(some_class2=SomeClass(expected_value))
    assert unparsed_args == shlex.split(f"--a {passed_value}")

    with raises_missing_required_arg("-some_class2.a/--some_class2.a"):
        parser.parse_known_args(shlex.split(f"--a {passed_value}"))

    with raises_missing_required_arg("-some_class2.a/--some_class2.a"):
        parser.parse_args("")

    assert parser.parse_args(shlex.split(f"--some_class2.a {passed_value}")) == argparse.Namespace(
        some_class2=SomeClass(expected_value)
    )


def test_with_regular_suppressed_args(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type
        b: int = 5

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default=argparse.SUPPRESS)
    if some_type == bool:
        parser.add_argument("--c", type=str2bool, default=argparse.SUPPRESS)
    else:
        parser.add_argument("--c", type=some_type, default=argparse.SUPPRESS)

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value} --b 0 --c {passed_value}"))
    assert parsed_args == argparse.Namespace(
        some_class={"a": expected_value, "b": 0}, c=expected_value
    )

    parsed_args, unparsed_args = parser.parse_known_args(shlex.split(f"--a {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": expected_value})
    assert unparsed_args == []

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value} --c {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": expected_value}, c=expected_value)

    assert parser.parse_args("") == argparse.Namespace()
    assert parser.parse_args(shlex.split("--b 0")) == argparse.Namespace(some_class={"b": 0})


def test_suppress_with_regular_unsuppressed_args(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type
        b: int = 5

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default=argparse.SUPPRESS)
    if some_type == bool:
        parser.add_argument("--c", type=str2bool)
    else:
        parser.add_argument("--c", type=some_type)

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value} --b 0 --c {passed_value}"))
    assert parsed_args == argparse.Namespace(
        some_class={"a": expected_value, "b": 0}, c=expected_value
    )

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": expected_value}, c=None)

    parsed_args = parser.parse_args(shlex.split(f"--a {passed_value} --c {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": expected_value}, c=expected_value)

    assert parser.parse_args("") == argparse.Namespace(c=None)
    assert parser.parse_args(shlex.split("--b 0")) == argparse.Namespace(
        some_class={"b": 0}, c=None
    )


def test_suppress_default_nested(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass()
    class SomeNestedClass:
        c: some_type

    @dataclass
    class SomeClass:
        a: SomeNestedClass
        b: int = 5

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default=argparse.SUPPRESS)

    parsed_args, unparsed_args = parser.parse_known_args(shlex.split(f"--c {passed_value} --b 0"))
    assert parsed_args == argparse.Namespace(some_class={"a": {"c": expected_value}, "b": 0})
    assert unparsed_args == []

    parsed_args, unparsed_args = parser.parse_known_args(shlex.split(f"--c {passed_value} --b 0"))
    assert parsed_args == argparse.Namespace(some_class={"a": {"c": expected_value}, "b": 0})
    assert unparsed_args == []

    parsed_args, unparsed_args = parser.parse_known_args(shlex.split(f"--c {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": {"c": expected_value}})
    assert unparsed_args == []

    assert parser.parse_args("") == argparse.Namespace()
    assert parser.parse_args(shlex.split("--b 0")) == argparse.Namespace(some_class={"b": 0})

    parsed_args = parser.parse_args(shlex.split(f"--c {passed_value}"))
    assert parsed_args == argparse.Namespace(some_class={"a": {"c": expected_value}})



import sys
import textwrap
from dataclasses import dataclass
from typing import Callable, ContextManager, List, Union

import pytest

from simple_parsing import helpers
from simple_parsing.helpers.fields import field, flag

from .testutils import (
    TestSetup,
    exits_and_writes_to_stderr,
    raises_expected_n_args,
    raises_missing_required_arg,
    raises_unrecognized_args,
)


@dataclass
class Base(TestSetup):
    """Some extension of base-class `Base`"""

    a: int = 5
    f: bool = False


@dataclass
class Flags(TestSetup):
    a: bool  # an example required flag (defaults to False)
    b: bool = True  # optional flag 'b'.
    c: bool = False  # optional flag 'c'.


@pytest.mark.parametrize(
    "flag,expected_f",
    [
        ("--a 5", False),
        ("--a 5 --f", True),
        ("--a 5 --nof", False),
        ("--a 5 --f=true", True),
        ("--a 5 --f true", True),
        ("--a 5 --f True", True),
        ("--a 5 --f=false", False),
        ("--a 5 --f false", False),
        ("--a 5 --f False", False),
        ("--a 5 --f --f false", False),
    ],
)
def test_bool_base_work(flag: str, expected_f: bool):
    assert Base.setup(flag).f is expected_f


def test_bool_args_in_help():
    prog = sys.argv[0].split("/")[-1]
    options = "options" if sys.version_info >= (3, 10) else "optional arguments"
    assert Flags.get_help_text() == textwrap.dedent(
        f"""\
        usage: {prog} [-h] -a bool [-b bool] [-c bool]

        {options}:
          -h, --help            show this help message and exit

        Flags ['flags']:
          Flags(a: bool, b: bool = True, c: bool = False)

          -a bool, --a bool, --noa bool
                                an example required flag (defaults to False) (default:
                                None)
          -b bool, --b bool, --nob bool
                                optional flag 'b'. (default: True)
          -c bool, --c bool, --noc bool
                                optional flag 'c'. (default: False)
        """
    )


def test_bool_attributes_work():
    ext = Base.setup("--a 5 --f")
    assert ext.f is True

    ext = Base.setup("--a 5")
    assert ext.f is False


@pytest.mark.parametrize(
    "flag,a,b,c",
    [
        ("--a", True, True, False),
        ("--a true --b --c", True, True, True),
        ("--a true --noc --b --c --noc", True, True, False),
        ("--noa --b false --noc", False, False, False),
    ],
)
def test_bool_flags_work(flag: str, a: bool, b: bool, c: bool):
    flags = Flags.setup(flag)
    assert flags.a is a
    assert flags.b is b
    assert flags.c is c


@pytest.mark.parametrize("invalid_nargs", [0, 1, "*", "+"])
def test_bool_field_doesnt_accept_invalid_nargs(invalid_nargs: Union[int, str]):
    """When using a `bool` annotation on a field, it doesn't accept `nargs` other than ? or None.

    This is because in argparse, `nargs=0` means a list with no values, and `nargs=1` is a list
    with one value (not just a required bool value).
    """

    @dataclass
    class Options(TestSetup):
        val: bool = field(default=True, nargs=invalid_nargs)

    with pytest.raises(ValueError, match="Invalid nargs for bool field"):
        Options.setup("")


@pytest.mark.parametrize(
    "flag, nargs, a_or_failure",
    [
        # By default, support both --noflag and --flag=false
        ("--a", "?", True),
        ("--noa", "?", False),
        ("--a true", "?", True),
        ("--a true false", "?", lambda: raises_unrecognized_args("false")),
        # `nargs=None` is like `nargs='?'`
        ("--a", None, True),
        ("--noa", None, False),
        ("--a true", None, True),
        ("--a true false", None, lambda: raises_unrecognized_args("false")),
    ],
)
def test_bool_nargs(
    flag,
    nargs,
    a_or_failure: Union[bool, Callable[[], ContextManager]],
):
    @dataclass
    class MyClass(TestSetup):
        """Some test class"""

        a: bool = helpers.field(nargs=nargs)

    if isinstance(a_or_failure, (bool, list, tuple)):
        a = a_or_failure
        flags = MyClass.setup(flag)
        assert flags.a == a
    else:
        expect_failure_context_manager = a_or_failure
        with expect_failure_context_manager():
            MyClass.setup(flag)


@pytest.mark.parametrize(
    "flag, nargs, a_or_failure",
    [
        # 1 argument explicitly required
        ("--a", 1, lambda: raises_expected_n_args(1)),
        ("--noa", 1, lambda: raises_missing_required_arg("-a/--a")),
        ("--a=true", 1, [True]),
        ("--a true false", 1, lambda: raises_unrecognized_args("false")),
        # 2 argument explicitly required
        ("--a", 2, lambda: raises_expected_n_args(2)),
        ("--noa", 2, lambda: raises_missing_required_arg("-a/--a")),
        ("--a=true", 2, lambda: raises_expected_n_args(2)),
        ("--a true false", 2, [True, False]),
        # 1+ argument explicitly required
        ("--a", "+", lambda: exits_and_writes_to_stderr("expected at least one argument")),
        ("--noa", "+", lambda: exits_and_writes_to_stderr("the following arguments are required")),
        ("--a=true", "+", [True]),
        ("--a true false", "+", [True, False]),
        # 0 or 1+ argument explicitly required
        ("--a", "*", []),
        ("--noa", "*", lambda: raises_unrecognized_args("--noa")),
        ("--a=true", "*", [True]),
        ("--a true false", "*", [True, False]),
    ],
)
def test_list_of_bools_nargs(
    flag,
    nargs,
    a_or_failure: Union[List[bool], Callable[[], ContextManager]],
):
    @dataclass
    class MyClass(TestSetup):
        """Some test class"""

        a: List[bool] = helpers.field(nargs=nargs)

    if isinstance(a_or_failure, (list, tuple)):
        a = a_or_failure
        flags = MyClass.setup(flag)
        assert flags.a == a
    else:
        expect_failure_context_manager = a_or_failure
        with expect_failure_context_manager():
            MyClass.setup(flag)


@pytest.mark.parametrize("field", [field, flag])
def test_using_custom_negative_prefix(field):
    """Check that we can customize the negative prefix for a boolean field, either with the
    `field` function or with the `flag` function.
    """

    @dataclass
    class Config(TestSetup):
        debug: bool = field(default=False, negative_prefix="--no-")

    help_text = Config.get_help_text()
    assert "--nodebug" not in help_text
    assert "--no-debug" in help_text
    assert Config.setup("--no-debug").debug is False

    @dataclass
    class OtherConfig(TestSetup):
        """Same as above but the default is `True` for `debug`."""

        debug: bool = field(default=True, negative_prefix="--no-debug")

    help_text = OtherConfig.get_help_text()
    assert "--nodebug" not in help_text
    assert "--no-debug" in help_text
    assert OtherConfig.setup("--no-debug").debug is False


a = flag(default=True, negative_prefix="--no-")


@pytest.mark.parametrize("field", [field, flag])
def test_using_custom_negative_option(field):
    from simple_parsing.helpers.fields import flag

    @dataclass
    class Config(TestSetup):
        debug: bool = field(default=False, negative_option="--release")

    help_text = Config.get_help_text()
    assert "--nodebug" not in help_text
    assert "--release" in help_text
    assert Config.setup("--release").debug is False

    @dataclass
    class OtherConfig(TestSetup):
        """Same as above but the default is `True` for `debug`."""

        debug: bool = flag(default=True, negative_option="--no-debug")

    help_text = OtherConfig.get_help_text()
    assert "--nodebug" not in help_text
    assert "--no-debug" in help_text
    assert OtherConfig.setup("--no-debug").debug is False


@pytest.mark.parametrize("default_value", [True, False])
def test_nested_bool_field_negative_args(default_value: bool):
    """Test that we get --train.nodebug instead of --notrain.debug"""

    @dataclass
    class Options:
        debug: bool = default_value

    @dataclass
    class Config(TestSetup):
        train: Options = field(default_factory=Options)
        valid: Options = field(default_factory=Options)

    assert Config.setup("") == Config()
    assert Config.setup("--train.debug") == Config(train=Options(debug=True))
    assert Config.setup("--train.nodebug") == Config(train=Options(debug=False))

    help_text = Config.get_help_text()
    assert "--notrain.debug" not in help_text
    assert "--novalid.debug" not in help_text
    assert "--valid.nodebug" in help_text
    assert "--train.nodebug" in help_text


@pytest.mark.parametrize("default_value", [True, False])
def test_using_abbrev_feature(default_value: bool):
    @dataclass
    class Options(TestSetup):
        verbose: bool = field(default=default_value, negative_option="silent")

    assert Options.setup("--verbose") == Options(verbose=True)
    assert Options.setup("--verb") == Options(verbose=True)
    assert Options.setup("--silent") == Options(verbose=False)
    assert Options.setup("--sil") == Options(verbose=False)


@pytest.mark.parametrize("default_value", [True, False])
def test_nested_bool_field_negative_option_conflict(default_value: bool):
    """Check that the negative option strings also get a prefix when there's a conflict."""

    @dataclass
    class Options:
        verbose: bool = field(default=default_value, negative_option="silent")

    @dataclass
    class Config(TestSetup):
        train: Options = field(default_factory=Options)
        valid: Options = field(default_factory=Options)

    help_text = Config.get_help_text()
    assert "--train.silent" in help_text
    assert "--valid.silent" in help_text
    assert "--silent" not in help_text


@pytest.mark.xfail(
    reason=(
        "TODO: When a field has a single letter name, we usually add `-{name}` and `--{name}`. "
        "However, if a prefix gets added when there's a conflict, then we'd like to only generate "
        "`--{prefix}.{name}`, and *not* `-{prefix}.{name}`."
    ),
    strict=True,
    raises=AssertionError,
)
@pytest.mark.parametrize("bool_field", [field, flag])
@pytest.mark.parametrize("default_value", [True, False])
def test_bool_nested_field_when_conflict_has_two_dashes(
    bool_field: Callable[..., bool], default_value: bool
):
    """TODO:"""
    # Check that there isn't a "-train.d bool" argument generated here, only "--train.d bool"
    @dataclass
    class Options:

        # whether or not to execute in debug mode.
        d: bool = bool_field(default=default_value)

    @dataclass
    class Config(TestSetup):
        train: Options = field(default_factory=Options)
        valid: Options = field(default_factory=Options)

    help_text = Config.get_help_text()
    assert "--train.d bool" in help_text
    assert " -train.d bool" not in help_text



from dataclasses import dataclass

import pytest

from simple_parsing.wrappers.field_wrapper import ArgumentGenerationMode, NestedMode

from . import TestSetup


@dataclass
class ModelOptions:
    path: str
    device: str


@dataclass
class ServerOptions(TestSetup):
    host: str
    port: int
    model: ModelOptions


expected = ServerOptions(host="myserver", port=80, model=ModelOptions(path="a_path", device="cpu"))


def test_flat():
    options = ServerOptions.setup(
        "--host myserver " "--port 80 " "--path a_path " "--device cpu",
    )
    assert options == expected

    with pytest.raises(SystemExit):
        ServerOptions.setup(
            "--opts.host myserver "
            "--opts.port 80 "
            "--opts.model.path a_path "
            "--opts.model.device cpu",
            dest="opts",
        )


@pytest.mark.parametrize("without_root", [True, False])
def test_both(without_root):
    options = ServerOptions.setup(
        "--host myserver " "--port 80 " "--path a_path " "--device cpu",
        dest="opts",
        argument_generation_mode=ArgumentGenerationMode.BOTH,
    )
    assert options == expected

    args = (
        "--opts.host myserver "
        "--opts.port 80 "
        "--opts.model.path a_path "
        "--opts.model.device cpu"
    )
    if without_root:
        args = args.replace("opts.", "")
    options = ServerOptions.setup(
        args,
        dest="opts",
        argument_generation_mode=ArgumentGenerationMode.BOTH,
        nested_mode=NestedMode.WITHOUT_ROOT if without_root else NestedMode.DEFAULT,
    )
    assert options == expected


@pytest.mark.parametrize("without_root", [True, False])
def test_nested(without_root):
    with pytest.raises(SystemExit):
        options = ServerOptions.setup(
            "--host myserver " "--port 80 " "--path a_path " "--device cpu",
            dest="opts",
            argument_generation_mode=ArgumentGenerationMode.NESTED,
        )

    args = (
        "--opts.host myserver "
        "--opts.port 80 "
        "--opts.model.path a_path "
        "--opts.model.device cpu"
    )
    if without_root:
        args = args.replace("opts.", "")
    options = ServerOptions.setup(
        args,
        dest="opts",
        argument_generation_mode=ArgumentGenerationMode.NESTED,
        nested_mode=NestedMode.WITHOUT_ROOT if without_root else NestedMode.DEFAULT,
    )
    assert options == expected



from dataclasses import dataclass
from test.testutils import TestSetup

from simple_parsing import field


def test_aliases_with_given_dashes():
    @dataclass
    class Foo(TestSetup):
        output_dir: str = field(default="/out", alias=["-o", "--out"])

    foo = Foo.setup("--output_dir /bob")
    assert foo.output_dir == "/bob"

    foo = Foo.setup("-o /cat")
    assert foo.output_dir == "/cat"

    foo = Foo.setup("--out /john")
    assert foo.output_dir == "/john"


def test_aliases_without_dashes():
    @dataclass
    class Foo(TestSetup):
        output_dir: str = field(default="/out", alias=["o", "out"])

    foo = Foo.setup("--output_dir /bob")
    assert foo.output_dir == "/bob"

    foo = Foo.setup("-o /cat")
    assert foo.output_dir == "/cat"

    foo = Foo.setup("--out /john")
    assert foo.output_dir == "/john"



from dataclasses import dataclass, field
from typing import Any, List, Tuple, Type

import pytest

from simple_parsing.helpers import list_field

from .testutils import (
    ListFormattingFunction,
    ListOfListsFormattingFunction,
    TestSetup,
    format_list_using_brackets,
    format_list_using_double_quotes,
    format_list_using_single_quotes,
    format_list_using_spaces,
    format_lists_using_brackets,
    format_lists_using_double_quotes,
    format_lists_using_single_quotes,
    parametrize,
    raises_missing_required_arg,
    xfail_param,
)


def test_list_one_element(simple_attribute):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class Container(TestSetup):
        a: List[some_type] = list_field()

    c = Container.setup("")
    assert c.a == []
    c = Container.setup(f"--a {passed_value}")
    assert c.a == [expected_value], Container.get_help_text()


@pytest.fixture
def ContainerClass():
    @dataclass
    class ContainerClass(TestSetup):
        a: Tuple[int]
        b: List[int]
        c: Tuple[str] = tuple()
        d: List[int] = list_field()

    return ContainerClass


def test_single_element_list(ContainerClass):
    container = ContainerClass.setup("--a 1 --b 4 --c 7 --d 10")
    assert container.a == (1,)
    assert container.b == [4]
    assert container.c == ("7",)
    assert container.d == [10]


def test_required_attributes_works(ContainerClass):
    with raises_missing_required_arg():
        ContainerClass.setup("--b 4")

    with raises_missing_required_arg():
        ContainerClass.setup("--a 4")

    container = ContainerClass.setup("--a 4 --b 5")
    assert container == ContainerClass(a=(4,), b=[5])


def test_default_value(ContainerClass):
    container = ContainerClass.setup("--a 1 --b 4 5 6")
    assert container.a == (1,)
    assert container.b == [4, 5, 6]
    assert container.c == tuple()
    assert container.d == list()


@parametrize(
    "list_formatting_function",
    [
        format_list_using_spaces,
        # format_list_using_brackets,
        xfail_param(
            format_list_using_brackets,
            reason="TODO: decide which syntax we want to allow for single lists.",
        ),
        xfail_param(
            format_list_using_double_quotes,
            reason="TODO: decide which syntax we want to allow for single lists.",
        ),
        xfail_param(
            format_list_using_single_quotes,
            reason="TODO: decide which syntax we want to allow for single lists.",
        ),
    ],
)
@parametrize(
    "item_type, passed_values",
    [
        (int, [1, 2]),
        (float, [1.1, 2.1]),
        (str, ["a", "b"]),
        (bool, [True, True]),
    ],
)
def test_list_supported_formats(
    list_formatting_function: ListFormattingFunction,
    item_type: Type,
    passed_values: List[Any],
):
    @dataclass
    class SomeClass(TestSetup):
        a: List[item_type] = field(default_factory=list)  # type: ignore
        """some docstring for attribute 'a'"""

    arguments = "--a " + list_formatting_function(passed_values)
    print(arguments)

    some_class = SomeClass.setup(arguments)

    assert some_class.a == passed_values
    assert isinstance(some_class, SomeClass)
    assert len(some_class.a) == 2
    assert all(isinstance(v, item_type) for v in some_class.a)


@parametrize(
    "list_of_lists_formatting_function",
    [
        format_lists_using_brackets,
        format_lists_using_single_quotes,
        format_lists_using_double_quotes,
    ],
)
@parametrize(
    "item_type, passed_values",
    [
        (int, [[1, 2], [4, 5], [7, 8]]),
        (float, [[1.1, 2.1], [4.2, 5.2], [7.2, 8.2]]),
        (str, [["a", "b"], ["c", "d"], ["e", "f"]]),
        (bool, [[True, True], [True, False], [False, True]]),
    ],
)
def test_parse_multiple_with_list_attributes(
    list_of_lists_formatting_function: ListOfListsFormattingFunction,
    item_type: Type,
    passed_values: List[List[Any]],
):
    @dataclass
    class SomeClass(TestSetup):
        a: List[item_type] = field(default_factory=list)  # type: ignore
        """some docstring for attribute 'a'"""

    arguments = "--a " + list_of_lists_formatting_function(passed_values)
    classes = list(SomeClass.setup_multiple(3, arguments))
    assert classes == [SomeClass(a=value) for value in passed_values]



from __future__ import annotations

import json
import pickle
import warnings
from collections import OrderedDict
from dataclasses import MISSING, Field, dataclass, fields, is_dataclass
from functools import partial
from importlib import import_module
from itertools import chain
from logging import getLogger
from pathlib import Path
from types import ModuleType
from typing import IO, Any, Callable, ClassVar, TypeVar, Union

from simple_parsing.utils import (
    DataclassT,
    all_subclasses,
    get_args,
    get_forward_arg,
    is_optional,
)

from .decoding import decode_field, register_decoding_fn
from .encoding import SimpleJsonEncoder, encode

DumpFn = Callable[[Any, IO], None]
DumpsFn = Callable[[Any], str]
LoadFn = Callable[[IO], dict]
LoadsFn = Callable[[str], dict]

logger = getLogger(__name__)

D = TypeVar("D", bound="SerializableMixin")

try:
    import yaml

    def ordered_dict_constructor(loader: yaml.Loader, node: yaml.Node):
        # NOTE(ycho): `deep` has to be true for `construct_yaml_seq`.
        value = loader.construct_sequence(node, deep=True)
        return OrderedDict(*value)

    def ordered_dict_representer(dumper: yaml.Dumper, instance: OrderedDict) -> yaml.Node:
        # NOTE(ycho): nested list for compatibility with PyYAML's representer
        node = dumper.represent_sequence("OrderedDict", [list(instance.items())])
        return node

    yaml.add_representer(OrderedDict, ordered_dict_representer)
    yaml.add_constructor("OrderedDict", ordered_dict_constructor)
    yaml.add_constructor(
        "tag:yaml.org,2002:python/object/apply:collections.OrderedDict",
        ordered_dict_constructor,
    )

except ImportError:
    pass


class SerializableMixin:
    """Makes a dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ...
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {'a': 123, 'b': '456'}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    """

    subclasses: ClassVar[list[type[D]]] = []
    decode_into_subclasses: ClassVar[bool] = False

    def __init_subclass__(
        cls, decode_into_subclasses: bool | None = None, add_variants: bool = True
    ):
        logger.debug(f"Registering a new Serializable subclass: {cls}")
        super().__init_subclass__()
        if decode_into_subclasses is None:
            # if decode_into_subclasses is None, we will use the value of the
            # parent class, if it is also a subclass of Serializable.
            # Skip the class itself as well as object.
            parents = cls.mro()[1:-1]
            logger.debug(f"parents: {parents}")

            for parent in parents:
                if parent in SerializableMixin.subclasses and parent is not SerializableMixin:
                    decode_into_subclasses = parent.decode_into_subclasses
                    logger.debug(
                        f"Parent class {parent} has decode_into_subclasses = {decode_into_subclasses}"
                    )
                    break

        cls.decode_into_subclasses = decode_into_subclasses or False
        if cls not in SerializableMixin.subclasses:
            SerializableMixin.subclasses.append(cls)

        encode.register(cls, cls.to_dict)
        register_decoding_fn(cls, cls.from_dict)

    def to_dict(
        self, dict_factory: type[dict] = dict, recurse: bool = True, save_dc_types: bool = False
    ) -> dict:
        """Serializes this dataclass to a dict.

        NOTE: This 'extends' the `asdict()` function from
        the `dataclasses` package, allowing us to not include some fields in the
        dict, or to perform some kind of custom encoding (for instance,
        detaching `Tensor` objects before serializing the dataclass to a dict).
        """
        return to_dict(
            self, dict_factory=dict_factory, recurse=recurse, save_dc_types=save_dc_types
        )

    @classmethod
    def from_dict(cls: type[D], obj: dict, drop_extra_fields: bool | None = None) -> D:
        """Parses an instance of `cls` from the given dict.

        NOTE: If the `decode_into_subclasses` class attribute is set to True (or
        if `decode_into_subclasses=True` was passed in the class definition),
        then if there are keys in the dict that aren't fields of the dataclass,
        this will decode the dict into an instance the first subclass of `cls`
        which has all required field names present in the dictionary.

        Passing `drop_extra_fields=None` (default) will use the class attribute
        described above.
        Passing `drop_extra_fields=True` will decode the dict into an instance
        of `cls` and drop the extra keys in the dict.
        Passing `drop_extra_fields=False` forces the above-mentioned behaviour.
        """
        return from_dict(cls, obj, drop_extra_fields=drop_extra_fields)

    def dump(self, fp: IO[str], dump_fn: DumpFn = json.dump) -> None:
        dump(self, fp=fp, dump_fn=dump_fn)

    def dump_json(self, fp: IO[str], dump_fn: DumpFn = json.dump, **kwargs) -> None:
        return dump_json(self, fp, dump_fn=dump_fn, **kwargs)

    def dump_yaml(self, fp: IO[str], dump_fn: DumpFn | None = None, **kwargs) -> None:
        return dump_yaml(self, fp, dump_fn=dump_fn, **kwargs)

    def dumps(self, dump_fn: DumpsFn = json.dumps, **kwargs) -> str:
        return dumps(self, dump_fn=dump_fn, **kwargs)

    def dumps_json(self, dump_fn: DumpsFn = json.dumps, **kwargs) -> str:
        return dumps_json(self, dump_fn=dump_fn, **kwargs)

    def dumps_yaml(self, dump_fn: DumpsFn | None = None, **kwargs) -> str:
        return dumps_yaml(self, dump_fn=dump_fn, **kwargs)

    @classmethod
    def load(
        cls: type[D],
        path: Path | str | IO[str],
        drop_extra_fields: bool | None = None,
        load_fn: LoadFn | None = None,
        **kwargs,
    ) -> D:
        """Loads an instance of `cls` from the given file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[Path, str, IO[str]]): Path or Path string or open file.
            drop_extra_fields (bool, optional): Whether to drop extra fields or
                to decode the dictionary into the first subclass with matching
                fields. Defaults to None, in which case we use the value of
                `cls.decode_into_subclasses`.
                For more info, see `cls.from_dict`.
            load_fn (Callable, optional): Which loading function to use. Defaults
                to None, in which case we try to use the appropriate loading
                function depending on `path.suffix`:
                {
                    ".yml": yaml.safe_load,
                    ".yaml": yaml.safe_load,
                    ".json": json.load,
                    ".pth": torch.load,
                    ".pkl": pickle.load,
                }

        Raises:
            RuntimeError: If the extension of `path` is unsupported.

        Returns:
            D: An instance of `cls`.
        """
        return load(cls, path=path, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def _load(
        cls: type[D],
        fp: IO[str],
        drop_extra_fields: bool | None = None,
        load_fn: LoadFn = json.load,
        **kwargs,
    ) -> D:
        return load(cls, path=fp, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def load_json(
        cls: type[D],
        path: str | Path,
        drop_extra_fields: bool | None = None,
        load_fn: LoadFn = json.load,
        **kwargs,
    ) -> D:
        """Loads an instance from the corresponding json-formatted file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[str, Path]): Path to a json-formatted file.
            load_fn ([type], optional): Loading function to use. Defaults to json.load.

        Returns:
            D: an instance of the dataclass.
        """
        return load_json(cls, path, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def load_yaml(
        cls: type[D],
        path: str | Path,
        drop_extra_fields: bool | None = None,
        load_fn=None,
        **kwargs,
    ) -> D:
        """Loads an instance from the corresponding yaml-formatted file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[str, Path]): Path to a yaml-formatted file.
            load_fn ([type], optional): Loading function to use. Defaults to
                None, in which case `yaml.safe_load` is used.

        Returns:
            D: an instance of the dataclass.
        """
        return load_yaml(cls, path, load_fn=load_fn, drop_extra_fields=drop_extra_fields, **kwargs)

    def save(self, path: str | Path, dump_fn=None) -> None:
        save(self, path=path, dump_fn=dump_fn)

    def _save(self, path: str | Path, dump_fn: DumpFn = json.dump, **kwargs) -> None:
        save(self, path=path, dump_fn=partial(dump_fn, **kwargs))

    def save_yaml(self, path: str | Path, dump_fn: DumpFn | None = None, **kwargs) -> None:
        save_yaml(self, path, dump_fn=dump_fn, **kwargs)

    def save_json(self, path: str | Path, dump_fn=json.dump, **kwargs) -> None:
        save_json(self, path, dump_fn=dump_fn, **kwargs)

    @classmethod
    def loads(
        cls: type[D],
        s: str,
        drop_extra_fields: bool | None = None,
        load_fn: LoadsFn = json.loads,
    ) -> D:
        return loads(cls, s, drop_extra_fields=drop_extra_fields, load_fn=load_fn)

    @classmethod
    def loads_json(
        cls: type[D],
        s: str,
        drop_extra_fields: bool | None = None,
        load_fn=json.loads,
        **kwargs,
    ) -> D:
        return loads_json(
            cls, s, drop_extra_fields=drop_extra_fields, load_fn=partial(load_fn, **kwargs)
        )

    @classmethod
    def loads_yaml(
        cls: type[D],
        s: str,
        drop_extra_fields: bool | None = None,
        load_fn: LoadsFn | None = None,
        **kwargs,
    ) -> D:
        return loads_yaml(cls, s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)


@dataclass
class Serializable(SerializableMixin):
    """Makes a dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ...
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {'a': 123, 'b': '456'}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    """


@dataclass(frozen=True)
class FrozenSerializable(SerializableMixin):
    """Makes a (frozen) dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ...
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {'a': 123, 'b': '456'}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    """


@dataclass
class SimpleSerializable(SerializableMixin, decode_into_subclasses=True):
    pass


S = TypeVar("S", bound=SerializableMixin)


def get_serializable_dataclass_types_from_forward_ref(
    forward_ref: type, serializable_base_class: type[S] = SerializableMixin
) -> list[type[S]]:
    """Gets all the subclasses of `serializable_base_class` that have the same name as the argument of this forward reference annotation."""
    arg = get_forward_arg(forward_ref)
    potential_classes: list[type] = []
    for serializable_class in serializable_base_class.subclasses:
        if serializable_class.__name__ == arg:
            potential_classes.append(serializable_class)
    return potential_classes


T = TypeVar("T")


def load(
    cls: type[DataclassT],
    path: Path | str | IO,
    drop_extra_fields: bool | None = None,
    load_fn: LoadFn | None = None,
) -> DataclassT:
    """Loads an instance of `cls` from the given file.

    First, `load_fn` is used to get a potentially nested dictionary of python primitives from a
    file. Then, a decoding function is applied to each value, based on the type annotation of the
    corresponding field. Finally, the resulting dictionary is used to instantiate an instance of
    the dataclass `cls`.

    - string -> `load_fn` (json/yaml/etc) -> dict with "raw" python values -> decode -> \
        dict with constructor arguments -> `cls`(**dict) -> instance of `cls`

    NOTE: This does not save the types of the dataclass fields. This is usually not an issue, since
    we can recover the right type to use by looking at subclasses of the annotated type. However,
    in some cases (e.g. subgroups), it might be useful to save all the types of all the
    fields, in which case you should probably use something like `yaml.dump`, directly passing it
    the dataclass, instead of this.

    Args:
        cls (Type[D]): A dataclass type to load.
        path (Path | str): Path or Path string or open file.
        drop_extra_fields (bool, optional): Whether to drop extra fields or
            to decode the dictionary into the first subclass with matching
            fields. Defaults to None, in which case we use the value of
            `cls.decode_into_subclasses`.
            For more info, see `cls.from_dict`.
        load_fn ([type], optional): Which loading function to use. Defaults
            to None, in which case we try to use the appropriate loading
            function depending on `path.suffix`:
            {
                ".yml": yaml.safe_load,
                ".yaml": yaml.safe_load,
                ".json": json.load,
                ".pth": torch.load,
                ".pkl": pickle.load,
            }

    Raises:
        RuntimeError: If the extension of `path` is unsupported.

    Returns:
        D: An instance of `cls`.
    """
    if isinstance(path, str):
        path = Path(path)
    if load_fn is None and isinstance(path, Path):
        # Load a dict from the file.
        d = read_file(path)
    elif load_fn:
        with (path.open() if isinstance(path, Path) else path) as f:
            d = load_fn(f)
    else:
        raise ValueError(
            "A loading function must be passed, since we got an io stream, and the "
            "extension can't be retrieved."
        )
    # Convert the dict into an instance of the class.
    if drop_extra_fields is None and getattr(cls, "decode_into_subclasses", None) is not None:
        drop_extra_fields = not getattr(cls, "decode_into_subclasses")
    return from_dict(cls, d, drop_extra_fields=drop_extra_fields)


def load_json(
    cls: type[DataclassT],
    path: str | Path,
    drop_extra_fields: bool | None = None,
    load_fn: LoadFn = json.load,
    **kwargs,
) -> DataclassT:
    """Loads an instance from the corresponding json-formatted file.

    Args:
        cls (Type[D]): A dataclass type to load.
        path (Union[str, Path]): Path to a json-formatted file.
        load_fn ([type], optional): Loading function to use. Defaults to json.load.

    Returns:
        D: an instance of the dataclass.
    """
    return load(cls, path, drop_extra_fields=drop_extra_fields, load_fn=partial(load_fn, **kwargs))


def loads(
    cls: type[DataclassT],
    s: str,
    drop_extra_fields: bool | None = None,
    load_fn: LoadsFn = json.loads,
) -> DataclassT:
    d = load_fn(s)
    return from_dict(cls, d, drop_extra_fields=drop_extra_fields)


def loads_json(
    cls: type[DataclassT],
    s: str,
    drop_extra_fields: bool | None = None,
    load_fn: LoadsFn = json.loads,
    **kwargs,
) -> DataclassT:
    return loads(cls, s, drop_extra_fields=drop_extra_fields, load_fn=partial(load_fn, **kwargs))


def loads_yaml(
    cls: type[DataclassT],
    s: str,
    drop_extra_fields: bool | None = None,
    load_fn: LoadsFn | None = None,
    **kwargs,
) -> DataclassT:
    import yaml

    load_fn = load_fn or yaml.safe_load
    return loads(cls, s, drop_extra_fields=drop_extra_fields, load_fn=partial(load_fn, **kwargs))


extensions_to_loading_fn: dict[str, Callable[[IO], Any]] = {
    ".json": json.load,
    ".pkl": pickle.load,
}
extensions_to_read_mode: dict[str, str] = {".pkl": "rb"}

extensions_to_write_mode: dict[str, str] = {".pkl": "wb"}
extensions_to_dump_fn: dict[str, Callable[[Any, IO], None]] = {
    ".json": json.dump,
    ".pkl": pickle.dump,
}
try:
    import yaml

    extensions_to_loading_fn[".yaml"] = yaml.safe_load
    extensions_to_loading_fn[".yml"] = yaml.safe_load
    extensions_to_dump_fn[".yaml"] = yaml.dump
    extensions_to_dump_fn[".yml"] = yaml.dump


except ImportError:
    pass

try:
    import numpy  # type: ignore

    extensions_to_loading_fn[".npy"] = numpy.load
    extensions_to_dump_fn[".npy"] = numpy.save
    extensions_to_read_mode[".npy"] = "rb"
    extensions_to_write_mode[".npy"] = "wb"

except ImportError:
    pass

try:
    import torch  # type: ignore

    extensions_to_loading_fn[".pth"] = torch.load
    extensions_to_dump_fn[".pth"] = torch.save
    extensions_to_read_mode[".pth"] = "rb"
    extensions_to_write_mode[".pth"] = "wb"
except ImportError:
    pass


def read_file(path: str | Path) -> dict:
    """Returns the contents of the given file as a dictionary.
    Uses the right function depending on `path.suffix`:
    {
        ".yml": yaml.safe_load,
        ".yaml": yaml.safe_load,
        ".json": json.load,
        ".pth": torch.load,
        ".pkl": pickle.load,
    }
    """
    path = Path(path)
    if path.suffix in extensions_to_loading_fn:
        load_fn = extensions_to_loading_fn[path.suffix]
    else:
        raise RuntimeError(
            f"Unable to determine what function to use in order to load "
            f"path {path} into a dictionary since the path's extension isn't registered in the "
            f"`extensions_to_loading_fn` dictionary..."
        )
    mode = extensions_to_read_mode.get(path.suffix, "r")
    with open(path, mode=mode) as f:
        return load_fn(f)


def save(
    obj: Any,
    path: str | Path,
    dump_fn: Callable[[dict, IO], None] | None = None,
    save_dc_types: bool = False,
) -> None:
    """Save the given dataclass or dictionary to the given file.

    Note: The `encode` function is applied to all the object fields to get serializable values,
    like so:
    - obj -> encode -> "raw" values (dicts, strings, ints, etc) -> `dump_fn` ([json/yaml/etc].dumps) -> string
    """
    path = Path(path)

    if not isinstance(obj, dict):
        obj = to_dict(obj, save_dc_types=save_dc_types)

    if dump_fn:
        save_fn = dump_fn
    elif path.suffix in extensions_to_dump_fn:
        save_fn = extensions_to_dump_fn[path.suffix]
    else:
        raise RuntimeError(
            f"Unable to determine what function to use in order to save obj {obj} to path {path},"
            f"since the path's extension isn't registered in the "
            f"`extensions_to_dump_fn` dictionary..."
        )
    mode = extensions_to_write_mode.get(path.suffix, "w")
    with open(path, mode=mode) as f:
        return save_fn(obj, f)


def save_yaml(
    obj, path: str | Path, dump_fn: DumpFn | None = None, save_dc_types: bool = False, **kwargs
) -> None:
    import yaml

    if dump_fn is None:
        dump_fn = yaml.dump
    save(obj, path, dump_fn=partial(dump_fn, **kwargs), save_dc_types=save_dc_types)


def save_json(
    obj, path: str | Path, dump_fn: DumpFn = json.dump, save_dc_types: bool = False, **kwargs
) -> None:
    save(obj, path, dump_fn=partial(dump_fn, **kwargs), save_dc_types=save_dc_types)


def load_yaml(
    cls: type[T],
    path: str | Path,
    drop_extra_fields: bool | None = None,
    load_fn: LoadFn | None = None,
    **kwargs,
) -> T:
    """Loads an instance from the corresponding yaml-formatted file.

    Args:
        cls (Type[T]): A dataclass type to load.
        path (Union[str, Path]): Path to a yaml-formatted file.
        load_fn ([type], optional): Loading function to use. Defaults to
            None, in which case `yaml.safe_load` is used.

    Returns:
        T: an instance of the dataclass.
    """
    import yaml

    if load_fn is None:
        load_fn = yaml.safe_load
    return load(cls, path, drop_extra_fields=drop_extra_fields, load_fn=partial(load_fn, **kwargs))


def dump(dc, fp: IO[str], dump_fn: DumpFn = json.dump) -> None:
    # Convert `dc` into a dict if needed.
    if not isinstance(dc, dict):
        dc = to_dict(dc)
    # Serialize that dict to the file using dump_fn.
    dump_fn(dc, fp)


def dump_json(dc, fp: IO[str], dump_fn: DumpFn = json.dump, **kwargs) -> None:
    return dump(dc, fp, dump_fn=partial(dump_fn, **kwargs))


def dump_yaml(dc, fp: IO[str], dump_fn: DumpFn | None = None, **kwargs) -> None:
    import yaml

    if dump_fn is None:
        dump_fn = yaml.dump
    return dump(dc, fp, dump_fn=partial(dump_fn, **kwargs))


def dumps(dc, dump_fn: DumpsFn = json.dumps) -> str:
    if not isinstance(dc, dict):
        dc = to_dict(dc)
    return dump_fn(dc)


def dumps_json(dc, dump_fn: DumpsFn = json.dumps, **kwargs) -> str:
    kwargs.setdefault("cls", SimpleJsonEncoder)
    return dumps(dc, dump_fn=partial(dump_fn, **kwargs))


def dumps_yaml(dc, dump_fn: DumpsFn | None = None, **kwargs) -> str:
    import yaml

    if dump_fn is None:
        dump_fn = yaml.dump
    return dumps(dc, dump_fn=partial(dump_fn, **kwargs))


DC_TYPE_KEY = "_type_"


def to_dict(
    dc: DataclassT,
    dict_factory: type[dict] = dict,
    recurse: bool = True,
    save_dc_types: bool = False,
) -> dict:
    """Serializes this dataclass to a dict.

    NOTE: This 'extends' the `asdict()` function from
    the `dataclasses` package, allowing us to not include some fields in the
    dict, or to perform some kind of custom encoding (for instance,
    detaching `Tensor` objects before serializing the dataclass to a dict).

    When `save_dc_types` is True, the type of each dataclass field is saved in the dict of that
    field at a `DC_TYPE_KEY` entry.
    """
    if not is_dataclass(dc):
        raise ValueError("to_dict should only be called on a dataclass instance.")

    d: dict[str, Any] = dict_factory()

    if save_dc_types:
        class_name = dc.__class__.__qualname__
        module = type(dc).__module__
        if "<locals>" in class_name:
            # Don't save the type of function-scoped dataclasses.
            warnings.warn(
                RuntimeWarning(
                    f"Dataclass type {type(dc)} is defined in a function scope, which might cause "
                    f"issues when deserializing the containing dataclass. Refusing to save the "
                    f"type of this dataclass in the serialized dictionary."
                )
            )
        else:
            d[DC_TYPE_KEY] = module + "." + class_name

    for f in fields(dc):
        name = f.name
        value = getattr(dc, name)

        # Do not include in dict if some corresponding flag was set in metadata.
        include_in_dict = f.metadata.get("to_dict", True)
        if not include_in_dict:
            continue

        custom_encoding_fn = f.metadata.get("encoding_fn")
        if custom_encoding_fn:
            # Use a custom encoding function if there is one.
            d[name] = custom_encoding_fn(value)
            continue

        encoding_fn = encode
        # TODO: Make a variant of the serialization tests that use the static functions everywhere.
        if is_dataclass(value) and recurse:
            encoded = to_dict(
                value, dict_factory=dict_factory, recurse=recurse, save_dc_types=save_dc_types
            )
            logger.debug(f"Encoded dataclass field {name}: {encoded}")
        else:
            try:
                encoded = encoding_fn(value)
            except Exception as e:
                logger.error(
                    f"Unable to encode value {value} of type {type(value)}! Leaving it as-is. (exception: {e})"
                )
                encoded = value
        d[name] = encoded
    return d


def from_dict(
    cls: type[DataclassT], d: dict[str, Any], drop_extra_fields: bool | None = None
) -> DataclassT:
    """Parses an instance of the dataclass `cls` from the dict `d`.

    Args:
        cls (Type[Dataclass]): A `dataclass` type.
        d (Dict[str, Any]): A dictionary of `raw` values, obtained for example
            when deserializing a json file into an instance of class `cls`.
        drop_extra_fields (bool, optional): Whether or not to drop extra
            dictionary keys (dataclass fields) when encountered. There are three
            options:
            - True:
                The extra keys are dropped, and this function returns an
                instance of `cls`.
            - False:
                The extra keys (if any) are kept, and we search through the
                subclasses of `cls` for the first dataclass which has all the
                required fields.
            - None (default):
                `drop_extra_fields = not cls.decode_into_subclasses`.

    Raises:
        RuntimeError: If an error is encountered while instantiating the class.

    Returns:
        Dataclass: An instance of the dataclass `cls`.
    """
    if d is None:
        return None

    obj_dict: dict[str, Any] = d.copy()

    init_args: dict[str, Any] = {}
    non_init_args: dict[str, Any] = {}

    if DC_TYPE_KEY in obj_dict:
        target = obj_dict.pop(DC_TYPE_KEY)
        # module, dc_type = target.rsplit(".", 1)
        live_dc_type = _locate(target)
        # live_module = importlib.import_module(module)
        # live_dc_type = getattr(live_module, dc_type)
        return from_dict(live_dc_type, obj_dict, drop_extra_fields=drop_extra_fields)

    if drop_extra_fields is None:
        drop_extra_fields = not getattr(cls, "decode_into_subclasses", False)
        logger.debug("drop_extra_fields is None. Using cls attribute.")

        if cls in {Serializable, FrozenSerializable, SerializableMixin}:
            # Passing `Serializable` means that we want to find the right
            # subclass depending on the keys.
            # We set the value to False when `Serializable` is passed, since
            # we use this mechanism when we don't know which dataclass to use.
            logger.debug("cls is `SerializableMixin`, drop_extra_fields = False.")
            drop_extra_fields = False

    logger.debug(f"from_dict for {cls}, drop extra fields: {drop_extra_fields}")
    for field in fields(cls) if is_dataclass(cls) else []:
        name = field.name
        if name not in obj_dict:
            if (
                field.metadata.get("to_dict", True)
                and field.default is MISSING
                and field.default_factory is MISSING
            ):
                logger.warning(
                    f"Couldn't find the field '{name}' in the dict with keys " f"{list(d.keys())}"
                )
            continue

        raw_value = obj_dict.pop(name)
        field_value = decode_field(
            field, raw_value, containing_dataclass=cls, drop_extra_fields=drop_extra_fields
        )

        if field.init:
            init_args[name] = field_value
        else:
            non_init_args[name] = field_value

    extra_args = obj_dict

    # If there are arguments left over in the dict after taking all fields.
    if extra_args:
        if drop_extra_fields:
            logger.warning(f"Dropping extra args {extra_args}")
            extra_args.clear()

        else:
            # Use the first Serializable derived class that has all the required
            # fields.
            logger.debug(f"Missing field names: {extra_args.keys()}")

            # Find all the "registered" subclasses of `cls`. (from Serializable)
            derived_classes: list[type[DataclassT]] = []

            for subclass in all_subclasses(cls):
                if subclass is not cls:
                    derived_classes.append(subclass)
            logger.debug(f"All derived classes of {cls} available: {derived_classes}")

            # All the arguments that the dataclass should be able to accept in
            # its 'init'.
            req_init_field_names = set(chain(extra_args, init_args))

            # Sort the derived classes by their number of init fields, so that
            # we choose the first one with all the required fields.
            derived_classes.sort(key=lambda dc: len(get_init_fields(dc)))

            for child_class in derived_classes:
                logger.debug(f"child class: {child_class.__name__}, mro: {child_class.mro()}")
                child_init_fields: dict[str, Field] = get_init_fields(child_class)
                child_init_field_names = set(child_init_fields.keys())

                if child_init_field_names >= req_init_field_names:
                    # `child_class` is the first class with all required fields.
                    logger.debug(f"Using class {child_class} instead of {cls}")
                    return from_dict(child_class, d, drop_extra_fields=False)

    init_args.update(extra_args)
    try:
        instance = cls(**init_args)  # type: ignore
    except TypeError as e:
        # raise RuntimeError(f"Couldn't instantiate class {cls} using init args {init_args}.")
        raise RuntimeError(
            f"Couldn't instantiate class {cls} using init args {init_args.keys()}: {e}"
        )

    for name, value in non_init_args.items():
        logger.debug(f"Setting non-init field '{name}' on the instance.")
        setattr(instance, name, value)
    return instance


def get_init_fields(dataclass: type) -> dict[str, Field]:
    result: dict[str, Field] = {}
    for field in fields(dataclass):
        if field.init:
            result[field.name] = field
    return result


def get_first_non_None_type(optional_type: type | tuple[type, ...]) -> type | None:
    if not isinstance(optional_type, tuple):
        optional_type = get_args(optional_type)
    for arg in optional_type:
        if arg is not Union and arg is not type(None):  # noqa: E721
            logger.debug(f"arg: {arg} is not union? {arg is not Union}")
            logger.debug(f"arg is not type(None)? {arg is not type(None)}")
            return arg
    return None


def is_dataclass_or_optional_dataclass_type(t: type) -> bool:
    """Returns whether `t` is a dataclass type or an Optional[<dataclass type>]."""
    return is_dataclass(t) or (is_optional(t) and is_dataclass(get_args(t)[0]))


def _locate(path: str) -> Any:
    """
    COPIED FROM Hydra:
    https://github.com/facebookresearch/hydra/blob/f8940600d0ab5c695961ad83abd042ffe9458caf/hydra/_internal/utils.py#L614

    Locate an object by name or dotted path, importing as necessary.
    This is similar to the pydoc function `locate`, except that it checks for
    the module from the given path from back to front.
    """
    if path == "":
        raise ImportError("Empty path")

    parts = [part for part in path.split(".")]
    for part in parts:
        if not len(part):
            raise ValueError(
                f"Error loading '{path}': invalid dotstring."
                + "\nRelative imports are not supported."
            )
    assert len(parts) > 0
    part0 = parts[0]
    try:
        obj = import_module(part0)
    except Exception as exc_import:
        raise ImportError(
            f"Error loading '{path}':\n{repr(exc_import)}"
            + f"\nAre you sure that module '{part0}' is installed?"
        ) from exc_import
    for m in range(1, len(parts)):
        part = parts[m]
        try:
            obj = getattr(obj, part)
        except AttributeError as exc_attr:
            parent_dotpath = ".".join(parts[:m])
            if isinstance(obj, ModuleType):
                mod = ".".join(parts[: m + 1])
                try:
                    obj = import_module(mod)
                    continue
                except ModuleNotFoundError as exc_import:
                    raise ImportError(
                        f"Error loading '{path}':\n{repr(exc_import)}"
                        + f"\nAre you sure that '{part}' is importable from module '{parent_dotpath}'?"
                    ) from exc_import
                except Exception as exc_import:
                    raise ImportError(
                        f"Error loading '{path}':\n{repr(exc_import)}"
                    ) from exc_import
            raise ImportError(
                f"Error loading '{path}':\n{repr(exc_attr)}"
                + f"\nAre you sure that '{part}' is an attribute of '{parent_dotpath}'?"
            ) from exc_attr
    return obj



from dataclasses import dataclass
from typing import List

from simple_parsing import ArgumentParser

from .testutils import TestSetup, exits_and_writes_to_stderr


def test_repro_issue_96():
    @dataclass
    class Options(TestSetup):
        list_items: List[str]  # SOMETHING

    parser = ArgumentParser(add_option_string_dash_variants=True)
    parser.add_arguments(Options, dest="options")

    with exits_and_writes_to_stderr(match="the following arguments are required: --list_items"):
        assert Options.setup("")

    assert Options.setup("--list_items foo") == Options(list_items=["foo"])
    assert Options.setup("--list_items") == Options(list_items=[])



import textwrap
from dataclasses import dataclass
from io import StringIO
from test.testutils import assert_help_output_equals

import pytest

import simple_parsing
from simple_parsing.wrappers.field_wrapper import ArgumentGenerationMode


@dataclass
class JBuildRelease:
    id: int
    url: str
    docker_image: str


def test_issue_46(assert_equals_stdout):
    parser = simple_parsing.ArgumentParser()
    parser.add_argument("--run_id", type=str)
    parser.add_arguments(JBuildRelease, dest="jbuild", prefix="jbuild")

    s = StringIO()
    parser.print_help(s)
    s.seek(0)
    output = str(s.read())

    assert_help_output_equals(
        actual=output,
        expected=textwrap.dedent(
            """\
        usage: pytest [-h] [--run_id str] --jbuildid int --jbuildurl str
                    --jbuilddocker_image str

        optional arguments:
        -h, --help            show this help message and exit
        --run_id str

        JBuildRelease ['jbuild']:
        JBuildRelease(id:int, url:str, docker_image:str)

        --jbuildid int
        --jbuildurl str
        --jbuilddocker_image str
        """
        ),
    )

    # assert_equals_stdout(
    #     textwrap.dedent(
    #         """\
    #         usage: pytest [-h] [--run_id str] --jbuildid int --jbuildurl str
    #                     --jbuilddocker_image str

    #         optional arguments:
    #         -h, --help            show this help message and exit
    #         --run_id str

    #         JBuildRelease ['jbuild']:
    #         JBuildRelease(id:int, url:str, docker_image:str)

    #         --jbuildid int
    #         --jbuildurl str
    #         --jbuilddocker_image str
    #         """
    #     )
    # )
    from .testutils import raises_missing_required_arg

    with raises_missing_required_arg():
        parser.parse_args(
            "--id 123 --jbuild.id 456 --jbuild.url bob --jbuild.docker_image foo".split()
        )


def test_issue_46_solution2(assert_equals_stdout):
    # This (now) works:
    parser = simple_parsing.ArgumentParser(argument_generation_mode=ArgumentGenerationMode.BOTH)
    parser.add_argument("--run_id", type=str)
    parser.add_arguments(JBuildRelease, dest="jbuild", prefix="jbuild.")
    s = StringIO()
    parser.print_help(s)
    s.seek(0)
    output = str(s.read())
    assert_help_output_equals(
        actual=output,
        expected=textwrap.dedent(
            """\
            usage: pytest [-h] [--run_id str] --jbuild.id int --jbuild.url str
                        --jbuild.docker_image str

            optional arguments:
            -h, --help            show this help message and exit
            --run_id str

            JBuildRelease ['jbuild']:
            JBuildRelease(id:int, url:str, docker_image:str)

            --jbuild.id int
            --jbuild.url str
            --jbuild.docker_image str
            """
        ),
    )


@pytest.mark.xfail(reason="TODO: Issue #49")
def test_conflict_with_regular_argparse_arg():
    # This _should_ work, but it doesn't, adding a new issue for this:
    # the problem: SimpleParsing doesn't yet detect
    # conflicts between arguments added the usual way with `add_argument` and those
    # added through `add_arguments`.
    parser = simple_parsing.ArgumentParser()
    parser.add_argument("--id", type=str)
    parser.add_arguments(JBuildRelease, dest="jbuild")
    args = parser.parse_args(
        "--id 123 --jbuild.id 456 --jbuild.url bob --jbuild.docker_image foo".split()
    )
    assert args.id == 123
    assert args.jbuild.id == 456


@pytest.mark.xfail(reason="TODO: Issue #49")
def test_workaround():
    pass

    # This also doesn't work, since the prefix is only added to the 'offending'
    # argument, rather than to all the args in that group.
    @dataclass
    class Main:
        id: int
        jbuild: JBuildRelease

    parser = simple_parsing.ArgumentParser()
    parser.add_arguments(Main, "main")
    args = parser.parse_args(
        "--id 123 --jbuild.id 456 --jbuild.url bob --jbuild.docker_image foo".split()
    )
    args = args.main
    assert args.id == 123
    assert args.jbuild.id == 456



""" Functions for decoding dataclass fields from "raw" values (e.g. from json).
"""
from __future__ import annotations

import inspect
import sys
import warnings
from collections import OrderedDict
from collections.abc import Mapping
from dataclasses import Field
from enum import Enum
from functools import partial
from logging import getLogger
from pathlib import Path
from typing import Any, Callable, TypeVar

from simple_parsing.annotation_utils.get_field_annotations import (
    evaluate_string_annotation,
)
from simple_parsing.utils import (
    get_bound,
    get_forward_arg,
    get_type_arguments,
    is_dataclass_type,
    is_dict,
    is_enum,
    is_forward_ref,
    is_list,
    is_literal,
    is_set,
    is_tuple,
    is_typevar,
    is_union,
    str2bool,
)

logger = getLogger(__name__)

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# Dictionary mapping from types/type annotations to their decoding functions.
_decoding_fns: dict[type[T], Callable[[Any], T]] = {
    # the 'primitive' types are decoded using the type fn as a constructor.
    t: t
    for t in [str, bytes]
}


def register_decoding_fn(
    some_type: type[T], function: Callable[[Any], T], overwrite: bool = False
) -> None:
    """Register a decoding function for the type `some_type`."""
    _register(some_type, function, overwrite=overwrite)


def _register(t: type, func: Callable, overwrite: bool = False) -> None:
    if t not in _decoding_fns or overwrite:
        # logger.debug(f"Registering the type {t} with decoding function {func}")
        _decoding_fns[t] = func


C = TypeVar("C", bound=Callable[[Any], Any])


def decoding_fn_for_type(some_type: type) -> Callable[[C], C]:
    """Registers a function to be used to convert a serialized value to the given type.

    The function should accept one argument (the serialized value) and return the decoded value.
    """

    def _wrapper(fn: C) -> C:
        register_decoding_fn(some_type, fn, overwrite=True)
        return fn

    return _wrapper


@decoding_fn_for_type(int)
def _decode_int(v: str) -> int:
    int_v = int(v)
    if isinstance(v, bool):
        warnings.warn(UnsafeCastingWarning(raw_value=v, decoded_value=int_v))
    elif int_v != float(v):
        warnings.warn(UnsafeCastingWarning(raw_value=v, decoded_value=int_v))
    return int_v


@decoding_fn_for_type(float)
def _decode_float(v: Any) -> float:
    float_v = float(v)
    if isinstance(v, bool):
        warnings.warn(UnsafeCastingWarning(raw_value=v, decoded_value=float_v))
    return float_v


@decoding_fn_for_type(bool)
def _decode_bool(v: Any) -> bool:
    if isinstance(v, str):
        bool_v = str2bool(v)
    else:
        bool_v = bool(v)
        if isinstance(v, (int, float)) and v not in (0, 1, 0.0, 1.0):
            warnings.warn(UnsafeCastingWarning(raw_value=v, decoded_value=bool_v))
    return bool_v


def decode_field(
    field: Field,
    raw_value: Any,
    containing_dataclass: type | None = None,
    drop_extra_fields: bool | None = None,
) -> Any:
    """Converts a "raw" value (e.g. from json file) to the type of the `field`.

    When serializing a dataclass to json, all objects are converted to dicts.
    The values which have a special type (not str, int, float, bool) are
    converted to string or dict. Hence this function allows us to recover the
    original type of pretty much any field which is of type `Serializable`, or
    has a registered decoding function (either through `register_decoding_fn` or
    through having `decoding_fn` passed directly to the `field` function.

    Args:
        field (Field): `Field` object from a dataclass.
        raw_value (Any): The `raw` value from deserializing the dataclass.

    Returns:
        Any: The "raw" value converted to the right type.
    """
    name = field.name
    field_type = field.type
    logger.debug(f"name = {name}, field_type = {field_type}")

    # If the user set a custom decoding function, we use it.
    custom_decoding_fn = field.metadata.get("decoding_fn")
    if custom_decoding_fn is not None:
        return custom_decoding_fn(raw_value)

    if isinstance(field_type, str) and containing_dataclass:
        field_type = evaluate_string_annotation(field_type, containing_dataclass)

    decoding_function = get_decoding_fn(field_type)

    _kwargs = dict(category=UnsafeCastingWarning) if sys.version_info >= (3, 11) else {}

    with warnings.catch_warnings(record=True, **_kwargs) as warning_messages:
        if is_dataclass_type(field_type) and drop_extra_fields is not None:
            # Pass the drop_extra_fields argument to the decoding function.
            decoded_value = decoding_function(raw_value, drop_extra_fields=drop_extra_fields)
        else:
            decoded_value = decoding_function(raw_value)

    for warning_message in warning_messages.copy():
        if not isinstance(warning_message.message, UnsafeCastingWarning):
            warnings.warn_explicit(
                message=warning_message.message,
                category=warning_message.category,
                filename=warning_message.filename,
                lineno=warning_message.lineno,
                # module=warning_message.module,
                # registry=warning_message.registry,
                # module_globals=warning_message.module_globals,
            )
            warning_messages.remove(warning_message)

    if warning_messages:
        warnings.warn(
            RuntimeWarning(
                f"Unsafe casting occurred when deserializing field '{name}' of type {field_type}: "
                f"raw value: {raw_value!r}, decoded value: {decoded_value!r}."
            )
        )
    return decoded_value


# NOTE: Disabling the caching here might help avoid some bugs, and it's unclear if this has that
# much of a performance impact.
def get_decoding_fn(type_annotation: type[T] | str) -> Callable[..., T]:
    """Fetches/Creates a decoding function for the given type annotation.

    This decoding function can then be used to create an instance of the type
    when deserializing dicts (which could have been obtained with JSON or YAML).

    This function inspects the type annotation and creates the right decoding
    function recursively in a "dynamic-programming-ish" fashion.
    NOTE: We cache the results in a `functools.lru_cache` decorator to avoid
    wasteful calls to the function. This makes this process pretty efficient.

    Args:
        t (Type[T]):
            A type or type annotation. Can be arbitrarily nested.
            For example:
            - List[int]
            - Dict[str, Foo]
            - Tuple[int, str, Any],
            - Dict[Tuple[int, int], List[str]]
            - List[List[List[List[Tuple[int, str]]]]]
            - etc.

    Returns:
        Callable[[Any], T]:
            A function that decodes a 'raw' value to an instance of type `t`.

    """
    from .serializable import from_dict

    logger.debug(f"Getting the decoding function for {type_annotation!r}")

    if isinstance(type_annotation, str):
        # Check first if there are any matching registered decoding functions.
        # TODO: Might be better to actually use the scope of the field, right?
        matching_entries = {
            key: decoding_fn
            for key, decoding_fn in _decoding_fns.items()
            if (inspect.isclass(key) and key.__name__ == type_annotation)
        }
        if len(matching_entries) == 1:
            _, decoding_fn = matching_entries.popitem()
            return decoding_fn
        elif len(matching_entries) > 1:
            # Multiple decoding functions match the type. Can't tell.
            logger.warning(
                RuntimeWarning(
                    f"More than one potential decoding functions were found for types that match "
                    f"the string annotation {type_annotation!r}. This will simply try each one "
                    f"and return the first one that works."
                )
            )
            return try_functions(*(decoding_fn for _, decoding_fn in matching_entries.items()))
        else:
            # Try to evaluate the string annotation.
            t = evaluate_string_annotation(type_annotation)

    elif is_forward_ref(type_annotation):
        forward_arg: str = get_forward_arg(type_annotation)
        # Recurse until we've resolved the forward reference.
        return get_decoding_fn(forward_arg)

    else:
        t = type_annotation

    logger.debug(f"{type_annotation!r} -> {t!r}")

    # T should now be a type or one of the objects from the typing module.

    if t in _decoding_fns:
        # The type has a dedicated decoding function.
        return _decoding_fns[t]

    if is_dataclass_type(t):
        return partial(from_dict, t)

    if t is Any:
        logger.debug(f"Decoding an Any type: {t}")
        return no_op

    if is_dict(t):
        logger.debug(f"Decoding a Dict field: {t}")
        args = get_type_arguments(t)
        if len(args) != 2:
            args = (Any, Any)
        return decode_dict(*args)

    if is_set(t):
        logger.debug(f"Decoding a Set field: {t}")
        args = get_type_arguments(t)
        if len(args) != 1:
            args = (Any,)
        return decode_set(args[0])

    if is_tuple(t):
        logger.debug(f"Decoding a Tuple field: {t}")
        args = get_type_arguments(t)
        return decode_tuple(*args)

    if is_list(t):
        logger.debug(f"Decoding a List field: {t}")
        args = get_type_arguments(t)
        if not args:
            # Using a `List` or `list` annotation, so we don't know what do decode the
            # items into!
            args = (Any,)
        assert len(args) == 1
        return decode_list(args[0])

    if is_union(t):
        logger.debug(f"Decoding a Union field: {t}")
        args = get_type_arguments(t)
        return decode_union(*args)

    if is_enum(t):
        logger.debug(f"Decoding an Enum field: {t}")
        return decode_enum(t)

    if is_typevar(t):
        bound = get_bound(t)
        logger.debug(f"Decoding a typevar: {t}, bound type is {bound}.")
        if bound is not None:
            return get_decoding_fn(bound)

    if is_literal(t):
        logger.debug(f"Decoding a Literal field: {t}")
        possible_vals = get_type_arguments(t)
        return decode_literal(*possible_vals)

    # Unknown type.
    warnings.warn(
        UserWarning(
            f"Unable to find a decoding function for the annotation {t} (of type {type(t)}). "
            f"Will try to use the type as a constructor. Consider registering a decoding function "
            f"using `register_decoding_fn`, or posting an issue on GitHub. "
        )
    )
    return try_constructor(t)


def decode_optional(t: type[T]) -> Callable[[Any | None], T | None]:
    decode = get_decoding_fn(t)

    def _decode_optional(val: Any | None) -> T | None:
        return val if val is None else decode(val)

    return _decode_optional


def try_functions(*funcs: Callable[[Any], T]) -> Callable[[Any], T | Any]:
    """Tries to use the functions in succession, else returns the same value unchanged."""

    def _try_functions(val: Any) -> T | Any:
        e: Exception | None = None
        for func in funcs:
            try:
                return func(val)
            except Exception as ex:
                e = ex
        else:
            logger.debug(f"Couldn't parse value {val}, returning it as-is. (exception: {e})")
        return val

    return _try_functions


def decode_union(*types: type[T]) -> Callable[[Any], T | Any]:
    types_list = list(types)
    optional = type(None) in types_list

    # Partition the Union into None and non-None types.
    while type(None) in types_list:
        types_list.remove(type(None))

    decoding_fns: list[Callable[[Any], T]] = [
        decode_optional(t) if optional else get_decoding_fn(t) for t in types_list
    ]

    # TODO: We could be a bit smarter about the order in which we try the functions, but for now,
    # we just try the functions in the same order as the annotation, and return the result from the
    # first function that doesn't raise an exception.

    # Try using each of the non-None types, in succession. Worst case, return the value.
    return try_functions(*decoding_fns)


def decode_list(t: type[T]) -> Callable[[list[Any]], list[T]]:
    decode_item = get_decoding_fn(t)

    def _decode_list(val: list[Any]) -> list[T]:
        return [decode_item(v) for v in val]

    return _decode_list


def decode_tuple(*tuple_item_types: type[T]) -> Callable[[list[T]], tuple[T, ...]]:
    """Makes a parsing function for creating tuples.

    Can handle tuples with different item types, for instance:
    - `Tuple[int, Foo, str, float, ...]`.

    Returns:
        Callable[[List[T]], Tuple[T, ...]]: A parsing function for creating tuples.
    """
    # Get the decoding function for each item type
    has_ellipsis = False
    if Ellipsis in tuple_item_types:
        # TODO: This isn't necessary, the ellipsis will always be at index 1.
        ellipsis_index = tuple_item_types.index(Ellipsis)
        decoding_fn_index = ellipsis_index - 1
        decoding_fn = get_decoding_fn(tuple_item_types[decoding_fn_index])
        has_ellipsis = True
    else:
        decoding_fns = [get_decoding_fn(t) for t in tuple_item_types]
    # Note, if there are more values than types in the tuple type, then the
    # last type is used.

    def _decode_tuple(val: tuple[Any, ...]) -> tuple[T, ...]:
        if has_ellipsis:
            return tuple(decoding_fn(v) for v in val)
        else:
            return tuple(decoding_fns[i](v) for i, v in enumerate(val))

    return _decode_tuple


def decode_set(item_type: type[T]) -> Callable[[list[T]], set[T]]:
    """Makes a parsing function for creating sets with items of type `item_type`.

    Args:
        item_type (Type[T]): the type of the items in the set.

    Returns:
        Callable[[List[T]], Set[T]]: [description]
    """
    # Get the parse fn for a list of items of type `item_type`.
    parse_list_fn = decode_list(item_type)

    def _decode_set(val: list[Any]) -> set[T]:
        return set(parse_list_fn(val))

    return _decode_set


def decode_dict(K_: type[K], V_: type[V]) -> Callable[[list[tuple[Any, Any]]], dict[K, V]]:
    """Creates a decoding function for a dict type. Works with OrderedDict too.

    Args:
        K_ (Type[K]): The type of the keys.
        V_ (Type[V]): The type of the values.

    Returns:
        Callable[[List[Tuple[Any, Any]]], Dict[K, V]]: A function that parses a
            Dict[K_, V_].
    """
    decode_k = get_decoding_fn(K_)
    decode_v = get_decoding_fn(V_)

    def _decode_dict(val: dict[Any, Any] | list[tuple[Any, Any]]) -> dict[K, V]:
        result: dict[K, V] = {}
        if isinstance(val, list):
            result = OrderedDict()
            items = val
        elif isinstance(val, OrderedDict):
            # NOTE(ycho): Needed to propagate `OrderedDict` type
            result = OrderedDict()
            items = val.items()
        else:
            items = val.items()
        for k, v in items:
            k_ = decode_k(k)
            v_ = decode_v(v)
            result[k_] = v_
        return result

    return _decode_dict


def decode_enum(item_type: type[Enum]) -> Callable[[str], Enum]:
    """
    Creates a decoding function for an enum type.

    Args:
        item_type (Type[Enum]): the type of the items in the set.

    Returns:
        Callable[[str], Enum]: A function that returns the enum member for the given name.
    """

    def _decode_enum(val: str) -> Enum:
        return item_type[val]

    return _decode_enum


def decode_literal(*possible_vals: Any) -> Callable[[Any], Any]:
    """Creates a decoding function for a Literal type.

    Args:
        *possible_vals (Any): The permissible values for the Literal type.

    Returns:
        Callable[[Any], Any]: A function that checks if a given value is one of the
            permissible values for the Literal. If not, raises a TypeError.
    """

    def _decode_literal(val: Any) -> Any:
        if val not in possible_vals:
            raise TypeError(f"Expected one of {possible_vals} for Literal, got {val}")

        return val

    return _decode_literal


def no_op(v: T) -> T:
    """Decoding function that gives back the value as-is.

    Args:
        v ([Any]): Any value.

    Returns:
        [type]: The value unchanged.
    """
    return v


def try_constructor(t: type[T]) -> Callable[[Any], T | Any]:
    """Tries to use the type as a constructor. If that fails, returns the value as-is.

    Args:
        t (Type[T]): A type.

    Returns:
        Callable[[Any], Union[T, Any]]: A decoding function that might return nothing.
    """

    def constructor(val):
        if isinstance(val, Mapping):
            return t(**val)
        else:
            return t(val)

    return try_functions(constructor)


register_decoding_fn(Path, Path)


class UnsafeCastingWarning(RuntimeWarning):
    def __init__(self, raw_value: Any, decoded_value: Any) -> None:
        super().__init__()
        self.raw_value = raw_value
        self.decoded_value = decoded_value



import sys
from dataclasses import InitVar, dataclass
from typing import Any, List, Tuple, Type

import pytest
from typing_extensions import Literal

from .testutils import TestSetup


@pytest.mark.skipif(
    sys.version_info[:2] < (3, 8),
    reason="Before 3.8 `InitVar[tp] is InitVar` so it's impossible to retrieve field type",
)
@pytest.mark.parametrize(
    "tp, passed_value, expected",
    [
        (int, "1", 1),
        (float, "1.4", 1.4),
        (Tuple[int, float], "2 -1.2", (2, -1.2)),
        (List[str], "12 abc", ["12", "abc"]),
        (Literal[1, 2, 3, "4"], "1", 1),
        (Literal[1, 2, 3, "4"], "4", "4"),
    ],
)
def test_initvar(tp: Type[Any], passed_value: str, expected: Any) -> None:
    @dataclass
    class Foo(TestSetup):
        init_var: InitVar[tp]

        def __post_init__(self, init_var: tp) -> None:
            assert init_var == expected

    Foo.setup(f"--init_var {passed_value}")



import functools
from typing import Any, Generic, TypeVar

_T = TypeVar("_T")


class npartial(functools.partial, Generic[_T]):
    """Partial that also invokes partials in args and kwargs before feeding them to the function.

    Useful for creating nested partials, e.g.:


    >>> from dataclasses import dataclass, field
    >>> @dataclass
    ... class Value:
    ...    v: int = 0
    >>> @dataclass
    ... class ValueWrapper:
    ...    value: Value
    ...
    >>> from functools import partial
    >>> @dataclass
    ... class WithRegularPartial:
    ...    wrapped: ValueWrapper = field(
    ...        default_factory=partial(ValueWrapper, value=Value(v=123)),
    ...    )

    Here's the problem: This here is BAD! They both share the same instance of Value!

    >>> WithRegularPartial().wrapped.value is WithRegularPartial().wrapped.value
    True
    >>> @dataclass
    ... class WithNPartial:
    ...    wrapped: ValueWrapper = field(
    ...        default_factory=npartial(ValueWrapper, value=npartial(Value, v=123)),
    ...    )
    >>> WithNPartial().wrapped.value is WithNPartial().wrapped.value
    False

    This is fine now!
    """

    def __call__(self, *args: Any, **keywords: Any) -> _T:
        keywords = {**self.keywords, **keywords}
        args = self.args + args
        args = tuple(arg() if isinstance(arg, npartial) else arg for arg in args)
        keywords = {k: v() if isinstance(v, npartial) else v for k, v in keywords.items()}
        return self.func(*args, **keywords)



""" A Partial helper that can be used to add arguments for an arbitrary class or callable. """
from __future__ import annotations

import dataclasses
import functools
import inspect
import typing
from dataclasses import make_dataclass
from functools import lru_cache, singledispatch, wraps
from logging import getLogger as get_logger
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Sequence,
    _ProtocolMeta,
    cast,
    get_type_hints,
)

from typing_extensions import ParamSpec, TypeVar

import simple_parsing

__all__ = ["Partial", "adjust_default", "config_for", "infer_type_annotation_from_default"]

C = TypeVar("C", bound=Callable)
_P = ParamSpec("_P")
_T = TypeVar("_T", bound=Any)
_C = TypeVar("_C", bound=Callable[..., Any])

logger = get_logger(__name__)


@singledispatch
def adjust_default(default: Any) -> Any:
    """Used the adjust the default value of a parameter that we extract from the signature.

    IF in some libraries, the signature has a special default value, that we shouldn't use as the
    default, e.g. "MyLibrary.REQUIRED" or something, then a handler can be registered here to
    convert it to something else.

    For example, here's a fix for the `lr` param of the `torch.optim.SGD` optimizer, which has a
    weird annotation of `_RequiredParameter`:

    ```python
    from torch.optim.optimizer import _RequiredParameter

    @adjust_default.register(_RequiredParameter)
    def _(default: Any) -> Any:
        return dataclasses.MISSING
    ```
    """
    return default


_P = ParamSpec("_P")
_OutT = TypeVar("_OutT")


def _cache_when_possible(fn: Callable[_P, _OutT]) -> Callable[_P, _OutT]:
    """Makes `fn` behave like `functools.cache(fn)` when args are all hashable, else no change."""
    cached_fn = lru_cache(maxsize=None)(fn)

    def _all_hashable(args: tuple, kwargs: dict) -> bool:
        return all(isinstance(arg, Hashable) for arg in args) and all(
            isinstance(arg, Hashable) for arg in kwargs.values()
        )

    @wraps(fn)
    def _switch(*args: _P.args, **kwargs: _P.kwargs) -> _OutT:
        if _all_hashable(args, kwargs):
            hashable_kwargs = typing.cast(Dict[str, Hashable], kwargs)
            return cached_fn(*args, **hashable_kwargs)
        return fn(*args, **kwargs)

    return _switch


@_cache_when_possible
def config_for(
    cls: type[_T] | Callable[_P, _T],
    ignore_args: str | Sequence[str] = (),
    frozen: bool = True,
    **defaults,
) -> type[Partial[_T]]:
    """Create a dataclass that contains the arguments for the constructor of `cls`.

    Example:

    >>> import dataclasses
    >>> import simple_parsing as sp
    >>> class Adam:  # i.e. `torch.optim.Adam`, which we don't have installed in this example.
    ...     def __init__(self, params, lr=1e-3, betas=(0.9, 0.999)):
    ...         self.params = params
    ...         self.lr = lr
    ...         self.betas = betas
    ...     def __repr__(self) -> str:
    ...         return f"Adam(params={self.params}, lr={self.lr}, betas={self.betas})"
    ...
    >>> AdamConfig = sp.config_for(Adam, ignore_args="params")
    >>> parser = sp.ArgumentParser()
    >>> _ = parser.add_arguments(AdamConfig, dest="optimizer")


    >>> args = parser.parse_args(["--lr", "0.1", "--betas", "0.1", "0.2"])
    >>> args.optimizer
    AdamConfig(lr=0.1, betas=(0.1, 0.2))

    The return dataclass is a subclass of `functools.partial` that returns the `Adam` object:

    >>> isinstance(args.optimizer, functools.partial)
    True
    >>> dataclasses.is_dataclass(args.optimizer)
    True
    >>> args.optimizer(params=[1, 2, 3])
    Adam(params=[1, 2, 3], lr=0.1, betas=(0.1, 0.2))

    >>> parser.print_help()  # doctest: +SKIP
    usage: pytest [-h] [--lr float] [--betas float float]

    options:
      -h, --help           show this help message and exit

    AdamConfig ['optimizer']:
      Auto-Generated configuration dataclass for simple_parsing.helpers.partial.Adam

      --lr float
      --betas float float
    """
    if isinstance(ignore_args, str):
        ignore_args = (ignore_args,)
    else:
        ignore_args = tuple(ignore_args)

    assert isinstance(defaults, dict)

    signature = inspect.signature(cls)

    fields: list[tuple[str, type, dataclasses.Field]] = []

    class_annotations = get_type_hints(cls)

    class_docstring_help = _parse_args_from_docstring(cls.__doc__ or "")
    if inspect.isclass(cls):
        class_constructor_help = _parse_args_from_docstring(cls.__init__.__doc__ or "")
    else:
        class_constructor_help = {}

    for name, parameter in signature.parameters.items():
        default = defaults.get(name, parameter.default)
        if default is parameter.empty:
            default = dataclasses.MISSING
        default = adjust_default(default)

        if name in ignore_args:
            logger.debug(f"Ignoring argument {name}")
            continue

        if parameter.annotation is not inspect.Parameter.empty:
            field_type = parameter.annotation
        elif name in class_annotations:
            field_type = class_annotations[name]
        elif default is not dataclasses.MISSING:
            # Infer the type from the default value.
            field_type = infer_type_annotation_from_default(default)
        else:
            logger.warning(
                f"Don't know what the type of field '{name}' of class {cls} is! "
                f"Ignoring this argument."
            )
            continue

        class_help_entries = {v for k, v in class_docstring_help.items() if k.startswith(name)}
        init_help_entries = {v for k, v in class_constructor_help.items() if k.startswith(name)}
        help_entries = init_help_entries or class_help_entries
        if help_entries:
            help_str = help_entries.pop()
        else:
            help_str = ""

        if default is dataclasses.MISSING:
            field = simple_parsing.field(help=help_str, required=True)
            # insert since fields without defaults need to go first.
            fields.insert(0, (name, field_type, field))
            logger.debug(f"Adding required field: {fields[0]}")
        else:
            field = simple_parsing.field(default=default, help=help_str)
            fields.append((name, field_type, field))
            logger.debug(f"Adding optional field: {fields[-1]}")

    cls_name = _get_generated_config_class_name(cls)
    config_class = make_dataclass(cls_name=cls_name, bases=(Partial,), fields=fields, frozen=frozen)
    config_class._target_ = cls
    config_class.__doc__ = (
        f"Auto-Generated configuration dataclass for {cls.__module__}.{cls.__qualname__}\n"
        + (cls.__doc__ or "")
    )

    return config_class


@singledispatch
def infer_type_annotation_from_default(default: Any) -> Any | type:
    """Used when there is a default value, but no type annotation, to infer the type of field to
    create on the config dataclass.
    """
    if isinstance(default, (int, str, float, bool)):
        return type(default)
    if isinstance(default, tuple):
        return typing.Tuple[tuple(infer_type_annotation_from_default(d) for d in default)]
    if isinstance(default, list):
        if not default:
            return list
        # Assuming that all items have the same type.
        return typing.List[infer_type_annotation_from_default(default[0])]
    if isinstance(default, dict):
        if not default:
            return dict
    raise NotImplementedError(
        f"Don't know how to infer type annotation to use for default of {default}"
    )


def _parse_args_from_docstring(docstring: str) -> dict[str, str]:
    """Taken from `pytorch_lightning.utilities.argparse`."""
    arg_block_indent = None
    current_arg = ""
    parsed = {}
    for line in docstring.split("\n"):
        stripped = line.lstrip()
        if not stripped:
            continue
        line_indent = len(line) - len(stripped)
        if stripped.startswith(("Args:", "Arguments:", "Parameters:")):
            arg_block_indent = line_indent + 4
        elif arg_block_indent is None:
            continue
        elif line_indent < arg_block_indent:
            break
        elif line_indent == arg_block_indent:
            current_arg, arg_description = stripped.split(":", maxsplit=1)
            parsed[current_arg] = arg_description.lstrip()
        elif line_indent > arg_block_indent:
            parsed[current_arg] += f" {stripped}"
    return parsed


def _get_generated_config_class_name(target: type | Callable) -> str:
    if inspect.isclass(target):
        return target.__name__ + "Config"
    elif inspect.isfunction(target):
        return target.__name__ + "_config"
    raise NotImplementedError(target)


class _Partial(_ProtocolMeta):
    _target_: _C

    def __getitem__(cls, target: Callable[_P, _T]) -> type[Callable[_P, _T]]:
        # full_path = target.__module__ + "." + target.__qualname__
        # if full_path in _autogenerated_config_classes:
        #     return _autogenerated_config_classes[full_path]

        # TODO: Maybe we should make a distinction here between Partial[_T] and Partial[SomeClass?]
        # Create the config class.
        config_class = config_for(target)
        # Set it's module to be the one calling this, and set that class name in the globals of
        # the calling module? --> No, too hacky.

        # OR: Set the module to be simple_parsing.helpers.partial ?
        # TODO: What if we had the name of the class directly encode how to recreate the class?
        config_class.__module__ = __name__
        _autogenerated_config_classes[config_class.__qualname__] = config_class
        return config_class


_autogenerated_config_classes: dict[str, type] = {}


def __getattr__(name: str):
    """
    Getting an attribute on this module here will check for the autogenerated config class with that name.
    """
    if name in globals():
        return globals()[name]

    if name in _autogenerated_config_classes:
        return _autogenerated_config_classes[name]

    raise AttributeError(f"Module {__name__} has no attribute {name}")


class Partial(functools.partial, Generic[_T], metaclass=_Partial):
    def __new__(cls, __func: Callable[_P, _T] | None = None, *args: _P.args, **kwargs: _P.kwargs):
        _func = __func or cls._target_
        assert _func is not None
        return super().__new__(cls, _func, *args, **kwargs)

    def __call__(self: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> _T:
        constructor_kwargs = {
            field.name: getattr(self, field.name) for field in dataclasses.fields(self)
        }
        constructor_kwargs.update(**kwargs)
        # TODO: Use `nested_partial` as a base class? (to instantiate all the partials inside as
        # well?)
        self = cast(Partial, self)
        return type(self)._target_(*args, **constructor_kwargs)

    def __getattr__(self, name: str):
        if name in self.keywords:
            return self.keywords[name]



"""Simple, extendable mechanism for encoding pracitaclly anything to string.

Just register a new encoder for a given type like so:

from simple_parsing.helpers.serialization import encode
import numpy as np
@encode.register
def encode_ndarray(obj: np.ndarray) -> str:
    return obj.tostring()
"""
import copy
import json
from argparse import Namespace
from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from enum import Enum
from functools import singledispatch
from logging import getLogger
from os import PathLike
from typing import Any, Dict, Hashable, List, Set, Tuple, Union

logger = getLogger(__name__)


class SimpleJsonEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        return encode(o)


"""
# NOTE: This code is commented because of static typing check error.
# The problem is incompatibility of mypy and singledispatch.
# See mypy issues for more info:
# https://github.com/python/mypy/issues/8356
# https://github.com/python/mypy/issues/2904
# https://github.com/python/mypy/issues/9112#issuecomment-725316936

class Dataclass(Protocol):
    # see dataclasses.is_dataclass implementation with _FIELDS
    __dataclass_fields__: Dict[str, Field[Any]]


T = TypeVar("T", bool, int, None, str)


@overload
def encode(obj: Dataclass) -> Dict[str, Any]: ...

@overload
def encode(obj: Union[List[Any], Set[Any], Tuple[Any, ...]]) -> List[Any]:
    ...

@overload
def encode(obj: Mapping[Any, Any]) -> Dict[Any, Any]: ...

@overload
def encode(obj: T) -> T: ...
"""


@singledispatch
def encode(obj: Any) -> Any:
    """Encodes an object into a json/yaml-compatible primitive type.

    This called to convert field attributes when calling `to_dict()` on a
    `DictSerializable` instance (including JsonSerializable and YamlSerializable).

    This is used as the 'default' keyword argument to `json.dumps` and
    `json.dump`, and is called when an object is encountered that `json` doesn't
    know how to serialize.

    To register a type as JsonSerializable, you can just register a custom
    serialization function. (There should be no need to do it for dataclasses,
    since that is supported by this function), use @encode.register
    (see the docs for singledispatch).
    """
    try:
        if is_dataclass(obj):
            # logger.debug(f"encoding object {obj} of class {type(obj)}")
            d: Dict[str, Any] = dict()
            for field in fields(obj):
                value = getattr(obj, field.name)
                try:
                    d[field.name] = encode(value)
                except TypeError as e:
                    logger.error(f"Unable to encode field {field.name}: {e}")
                    raise e
            return d
        else:
            # logger.debug(f"Deepcopying object {obj} of type {type(obj)}")
            return copy.deepcopy(obj)
    except Exception as e:
        logger.debug(f"Cannot encode object {obj}: {e}")
        raise e


@encode.register(list)
@encode.register(tuple)
# @encode.register(Sequence) # Would also encompass `str!`
@encode.register(set)
def encode_list(obj: Union[List[Any], Set[Any], Tuple[Any, ...]]) -> List[Any]:
    # TODO: Here we basically say "Encode all these types as lists before serializing"
    # That's ok for JSON, but YAML can serialize stuff directly though.
    # TODO: Also, with this, we also need to convert back to the right type when
    # deserializing, which is totally doable for the fields of dataclasses,
    # but maybe not for other stuff.
    return list(map(encode, obj))


@encode.register(Mapping)
def encode_dict(obj: Mapping) -> Dict[Any, Any]:
    constructor = type(obj)
    result = constructor()
    for k, v in obj.items():
        k_ = encode(k)
        v_ = encode(v)
        if isinstance(k_, Hashable):
            result[k_] = v_
        else:
            # If the encoded key isn't "Hashable", then we store it as a list of tuples
            if isinstance(result, dict):
                result = list(result.items())
            result.append((k_, v_))
    return result

    return type(obj)((encode(k), encode(v)) for k, v in obj.items())


@encode.register(PathLike)
def encode_path(obj: PathLike) -> str:
    return obj.__fspath__()


@encode.register(Namespace)
def encode_namespace(obj: Namespace) -> Any:
    return encode(vars(obj))


@encode.register(Enum)
def encode_enum(obj: Enum) -> str:
    return obj.name



from dataclasses import dataclass
from typing import List

from simple_parsing import field
from simple_parsing.docstring import AttributeDocString, get_attribute_docstring

from .testutils import TestSetup


@dataclass
class Base:
    """A simple base-class example"""

    a: int  # TODO: finetune this

    """docstring for attribute 'a'"""

    b: float = 5.0  # inline comment on attribute 'b'

    c: str = ""
    """Multi
    Line
    Docstring for 'c'
    """


@dataclass
class Extended(Base):
    """Some extension of base-class `Base`"""

    # Comment above d)
    # its multiline, does it still work?
    d: int = 5
    """ docstring for 'd' in Extended. """

    # Comment above e, but with a line skipped

    e: float = -1  # *# comment on the side of e


def test_docstring_parsing_work_on_base():
    docstring = get_attribute_docstring(Base, "a")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "TODO: finetune this"
    assert docstring.docstring_below == "docstring for attribute 'a'"

    docstring = get_attribute_docstring(Base, "b")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "inline comment on attribute 'b'"
    assert docstring.docstring_below == ""

    docstring = get_attribute_docstring(Base, "c")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == ""
    assert docstring.docstring_below == "Multi\nLine\nDocstring for 'c'\n"


def test_docstring_parsing_works_on_extended():
    docstring = get_attribute_docstring(Extended, "a")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "TODO: finetune this"
    assert docstring.docstring_below == "docstring for attribute 'a'"

    docstring = get_attribute_docstring(Extended, "b")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "inline comment on attribute 'b'"
    assert docstring.docstring_below == ""

    docstring = get_attribute_docstring(Extended, "c")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == ""
    assert docstring.docstring_below == "Multi\nLine\nDocstring for 'c'\n"

    docstring = get_attribute_docstring(Extended, "d")
    assert docstring.comment_above == "Comment above d)\nits multiline, does it still work?"
    assert docstring.comment_inline == ""
    assert docstring.docstring_below == "docstring for 'd' in Extended."

    docstring = get_attribute_docstring(Extended, "e")
    assert docstring.comment_above == "Comment above e, but with a line skipped"
    assert docstring.comment_inline == "*# comment on the side of e"
    assert docstring.docstring_below == ""


def test_docstring_works_with_field_function():
    @dataclass
    class UniqueFoo(TestSetup):
        """Some class Foo"""

        # A sequence of tasks.
        task_sequence: List[str] = field(choices=["train", "test", "ood"])  # side
        """Below"""

    docstring = get_attribute_docstring(UniqueFoo, "task_sequence")
    assert docstring.comment_above == "A sequence of tasks."
    assert docstring.comment_inline == "side"
    assert docstring.docstring_below == "Below"


def test_docstrings_with_multiple_inheritance():
    """Test to reproduce issue 162: https://github.com/lebrice/SimpleParsing/issues/162"""

    @dataclass
    class Fooz:
        bar: int = 123  #: The bar property

    @dataclass
    class Baz:
        bat: int = 123  #: The bat property

    @dataclass
    class FooBaz(Fooz, Baz):
        foobaz: int = 123  #: The foobaz property

    assert get_attribute_docstring(FooBaz, "bar") == AttributeDocString(
        comment_inline=": The bar property"
    )
    assert get_attribute_docstring(FooBaz, "bat") == AttributeDocString(
        comment_inline=": The bat property"
    )
    assert get_attribute_docstring(FooBaz, "foobaz") == AttributeDocString(
        comment_inline=": The foobaz property"
    )


def test_weird_docstring_with_field_like():
    @dataclass
    class FooA:
        """
        @dataclass
        class weird:
            bar: int = 123  # WRONG DOCSTRING
        """

        bar: int = 123  # The bar property

    assert get_attribute_docstring(FooA, "bar") == AttributeDocString(
        comment_inline="The bar property"
    )


def test_docstring_builds_upon_bases():
    @dataclass
    class Base2(TestSetup):
        """
        # WRONG ABOVE
        bar: int = 333 # WRONG INLINE
        '''WRONG DOCSTRING'''
        """

        bar: int = 123  # inline
        """field docstring from base class"""

    @dataclass
    class FooB(Base2):
        # Above
        bar: int = 123  # The bar property

    assert get_attribute_docstring(FooB, "bar") == AttributeDocString(
        comment_inline="The bar property",
        comment_above="Above",
        docstring_below="field docstring from base class",
    )

    assert "field docstring from base class" in FooB.get_help_text()


def test_getdocstring_bug():
    @dataclass
    class HParams:
        batch_size: int = 32

    @dataclass
    class Parent(TestSetup):
        batch_size: int = 48
        child: HParams = field(default_factory=lambda: HParams())

    assert get_attribute_docstring(Parent, "child") == AttributeDocString()


def test_desc_from_cls_docstring():
    @dataclass
    class SomeClass:
        """Creates a new object.

        Parameters
        ----------
        batch_size : int, optional
            _description_, by default 32

        """

        # above
        batch_size: int = 32  # side
        """below"""

    assert get_attribute_docstring(SomeClass, "batch_size") == AttributeDocString(
        desc_from_cls_docstring="_description_, by default 32",
        comment_above="above",
        comment_inline="side",
        docstring_below="below",
    )


def test_help_takes_value_from_docstring():
    @dataclass
    class Args(TestSetup):
        """args.

        Attributes:
            user: Username
            verbose: Display logs
        """

        user: str
        verbose: bool = False

    assert get_attribute_docstring(Args, "verbose") == AttributeDocString(
        desc_from_cls_docstring="Display logs",
    )



"""Tests for weird conflicts.
"""
import argparse
import functools
from dataclasses import dataclass, field

from simple_parsing import ArgumentParser

from .testutils import TestSetup, raises


def test_arg_and_dataclass_with_same_name(silent):
    @dataclass
    class SomeClass:
        a: int = 1  # some docstring for attribute 'a'

    parser = ArgumentParser()
    parser.add_argument("--a", default=123)
    with raises(argparse.ArgumentError):
        parser.add_arguments(SomeClass, dest="some_class")
        parser.parse_args("")


def test_arg_and_dataclass_with_same_name_after_prefixing(silent):
    @dataclass
    class SomeClass:
        a: int = 1  # some docstring for attribute 'a'

    @dataclass
    class Parent:
        pre: SomeClass = field(default_factory=lambda: SomeClass)
        bla: SomeClass = field(default_factory=lambda: SomeClass)

    parser = ArgumentParser()
    parser.add_argument("--pre.a", default=123, type=int)
    with raises(argparse.ArgumentError):
        parser.add_arguments(Parent, dest="some_class")
        parser.parse_args("--pre.a 123 --pre.a 456".split())


def test_weird_hierarchy():
    @dataclass
    class Base:
        v: float = 0.0

    @dataclass
    class A(Base):
        pass

    @dataclass
    class B(Base):
        pass

    @dataclass
    class C(Base):
        pass

    @dataclass
    class Options:
        a: A = field(default_factory=functools.partial(A, 0.1))
        b: B = field(default_factory=functools.partial(B, 0.2))

    @dataclass
    class Settings(TestSetup):
        opt: Options = field(default_factory=Options)
        c: Base = field(default_factory=functools.partial(C, 0.3))

    opt = Settings.setup("")
    print(opt)


def test_parent_child_conflict():
    @dataclass
    class HParams:
        batch_size: int = 32

    @dataclass
    class Parent2(TestSetup):
        batch_size: int = 48
        child: HParams = field(default_factory=HParams)

    p: Parent2 = Parent2.setup()
    assert p.child.batch_size == 32
    assert p.batch_size == 48



from .decoding import *
from .encoding import *
from .serializable import (
    FrozenSerializable,
    Serializable,
    SerializableMixin,
    dump,
    dump_json,
    dump_yaml,
    dumps,
    dumps_json,
    dumps_yaml,
    from_dict,
    load,
    load_json,
    load_yaml,
    save,
    save_json,
    save_yaml,
    to_dict,
)

try:
    from .yaml_serialization import YamlSerializable
except ImportError:
    pass
JsonSerializable = Serializable



from __future__ import annotations

import functools
import logging
from dataclasses import dataclass, field

import pytest

from simple_parsing import replace
from simple_parsing.utils import Dataclass, DataclassT

logger = logging.getLogger(__name__)


@dataclass
class A:
    a: float = 0.0


@dataclass
class B:
    b: str = "bar"
    b_post_init: str = field(init=False)

    def __post_init__(self):
        self.b_post_init = self.b + "_post"


@dataclass
class UnionConfig:
    a_or_b: A | B = field(default_factory=A)


@dataclass
class WithOptional:
    optional_a: A | None = None


@dataclass
class Level1:
    level: int = 1
    name: str = "level1"


@dataclass
class Level2:
    level: int = 2
    name: str = "level2"
    prev: Level1 = field(default_factory=Level1)


@dataclass
class Level3:
    level: int = 3
    name: str = "level3"
    prev: Level2 = field(default_factory=functools.partial(Level2, name="level2_foo"))


@dataclass
class InnerPostInit:
    in_arg: float = 1.0
    in_arg_post: str = field(init=False)
    for_outer_post: str = "foo"

    def __post_init__(self):
        self.in_arg_post = str(self.in_arg)


@dataclass
class OuterPostInit:
    out_arg: int = 1
    out_arg_post: str = field(init=False)
    inner: InnerPostInit = field(default_factory=InnerPostInit)
    arg_post_on_inner: str = field(init=False)

    def __post_init__(self):
        self.out_arg_post = str(self.out_arg)
        self.arg_post_on_inner = self.inner.for_outer_post + "_outer"


@pytest.mark.parametrize(
    ("start", "changes", "expected"),
    [
        (A(), {"a": 2.0}, A(a=2.0)),
        (A(a=2.0), {"a": 0.0}, A()),
        (B(), {"b": "test"}, B(b="test")),
        (B(b="test1"), {"b": "bar"}, B(b="bar")),
        (
            UnionConfig(a_or_b=A(a=1.0)),
            {"a_or_b": B(b="bob")},
            UnionConfig(a_or_b=B(b="bob")),
        ),
        (
            UnionConfig(a_or_b=A(a=1.0)),
            {"a_or_b.a": 2.0},
            UnionConfig(a_or_b=A(a=2.0)),
        ),
        # NOTE: This looks like we're converting a dict to a dataclass, but we're just replacing
        # the only field on the `A` object.
        (
            WithOptional(optional_a=A(a=0)),
            {"optional_a": {"a": 123}},
            WithOptional(optional_a=A(a=123)),
        ),
        # Test dataclasses with a non-init field and a __post_init__ method:
        (
            InnerPostInit(),
            {"in_arg": 2.0},
            InnerPostInit(in_arg=2.0),
        ),
        (
            OuterPostInit(),
            {"out_arg": 2, "inner": {"in_arg": 3.0, "for_outer_post": "bar"}},
            OuterPostInit(out_arg=2, inner=(InnerPostInit(3.0, for_outer_post="bar"))),
        ),
        (
            OuterPostInit(),
            {"out_arg": 2, "inner.in_arg": 3.0, "inner.for_outer_post": "bar"},
            OuterPostInit(out_arg=2, inner=(InnerPostInit(3.0, for_outer_post="bar"))),
        ),
    ],
)
def test_replace(start: DataclassT, changes: dict, expected: DataclassT):
    actual = replace(start, **changes)
    assert actual == expected


@pytest.mark.parametrize(
    ("start", "changes", "expected"),
    [
        (
            Level1(),
            {"name": "level1_the_potato"},
            Level1(name="level1_the_potato"),
        ),
        (
            Level2(),
            {"name": "level2_bar"},
            Level2(name="level2_bar"),
        ),
        (
            Level2(),
            {"name": "level2_bar", "prev": {"name": "level1_good"}},
            Level2(name="level2_bar", prev=Level1(name="level1_good")),
        ),
        (
            Level2(),
            {"name": "level2_bar", "prev.name": "level1_good"},
            Level2(name="level2_bar", prev=Level1(name="level1_good")),
        ),
        (
            Level3(),
            {
                "name": "level3_greatest",
                "prev": {"name": "level2_greater", "prev": {"name": "level1_greate"}},
            },
            Level3(
                name="level3_greatest",
                prev=Level2(name="level2_greater", prev=Level1(name="level1_greate")),
            ),
        ),
        (
            Level3(),
            {
                "name": "level3_greatest",
                "prev.name": "level2_greater",
                "prev.prev.name": "level1_great",
            },
            Level3(
                name="level3_greatest",
                prev=Level2(name="level2_greater", prev=Level1(name="level1_great")),
            ),
        ),
    ],
)
@pytest.mark.parametrize("pass_dict_as_kwargs", [True, False])
def test_replace_nested_dataclasses(
    start: DataclassT, changes: dict, expected: DataclassT, pass_dict_as_kwargs: bool
):
    actual = replace(start, **changes) if pass_dict_as_kwargs else replace(start, changes)
    assert actual == expected


@pytest.mark.parametrize(
    ("start", "changes", "exception_type", "match"),
    [
        (
            UnionConfig(a_or_b=A(a=1.0)),
            {"a_or_b.b": "bob"},
            TypeError,
            "unexpected keyword argument 'b'",
        ),
        (
            A(a=123),
            {"not_a_field": 456},
            TypeError,
            "unexpected keyword argument 'not_a_field'",
        ),
        (
            A(a=123),
            {"a_bad_field": {"foo": "bar"}},
            TypeError,
            "unexpected keyword argument 'a_bad_field'",
        ),
        (
            WithOptional(optional_a=None),
            {"foo.bar": {"a": 123}},
            TypeError,
            "unexpected keyword argument 'foo'",
        ),
        pytest.param(
            WithOptional(optional_a=None),
            {"optional_a.a": 123},
            # NOTE: This probably to be fixed... looks counter-intuitive. it's an artifact of us
            # converting {"optional_a.a": 123} to {"optional_a": {"a": 123}}. However, there's no
            # easy way to fix these cases without inspecting the type annotations of the fields,
            # which adds a lot of complexity.
            TypeError,
            "<todo>",
            marks=pytest.mark.xfail(strict=True, reason="TODO: Potentially ambiguous."),
        ),
    ],
)
@pytest.mark.parametrize("pass_dict_as_kwargs", [True, False])
def test_replace_invalid(
    start: Dataclass,
    changes: dict,
    exception_type: type[Exception],
    match: str,
    pass_dict_as_kwargs: bool,
):
    with pytest.raises(exception_type, match=match):
        if pass_dict_as_kwargs:
            replace(start, **changes)
        else:
            replace(start, changes)



import textwrap
from dataclasses import dataclass
from io import StringIO
from test.testutils import assert_help_output_equals

import pytest

from simple_parsing import ArgumentParser


@dataclass
class Options:
    """These are the options"""

    foo: str = "aaa"  # Description
    bar: str = "bbb"


@pytest.mark.xfail(reason="Issue64 is solved below.")
def test_reproduce_issue64():

    parser = ArgumentParser("issue64")

    parser.add_arguments(Options, dest="options")

    # args = parser.parse_args(["--help"])

    s = StringIO()
    parser.print_help(file=s)
    s.seek(0)

    assert s.read() == textwrap.dedent(
        """\
        usage: issue64 [-h] [--foo str] [--bar str]

        optional arguments:
            -h, --help  show this help message and exit

        Options ['options']:
            These are the options

            --foo str   Description (default: aaa)
            --bar str
        """
    )


def test_vanilla_argparse_issue64():
    """This test shows that the ArgumentDefaultsHelpFormatter of argparse doesn't add
    the "(default: xyz)" if the 'help' argument isn't already passed!

    This begs the question: Should simple-parsing add a 'help' argument always, so that
    the formatter can then add the default string after?
    """
    import argparse

    parser = ArgumentParser("issue64", formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    group = parser.add_argument_group("Options ['options']", description="These are the options")
    group.add_argument("--foo", type=str, metavar="str", default="aaa", help="Description")
    group.add_argument("--bar", type=str, metavar="str", default="bbb")

    from io import StringIO

    s = StringIO()
    parser.print_help(file=s)
    s.seek(0)
    output = str(s.read())
    assert_help_output_equals(
        actual=output,
        expected=textwrap.dedent(
            """\
            usage: issue64 [-h] [--foo str] [--bar str]

            optional arguments:
                -h, --help  show this help message and exit

            Options ['options']:
                These are the options

                --foo str   Description (default: aaa)
                --bar str
            """
        ),
    )


def test_solved_issue64():
    """test that shows that Issue 64 is solved now, by adding a single space as the
    'help' argument, the help formatter can then add the "(default: bbb)" after the
    argument.
    """
    parser = ArgumentParser("issue64")
    parser.add_arguments(Options, dest="options")

    s = StringIO()
    parser.print_help(file=s)
    s.seek(0)
    output = str(s.read())
    assert_help_output_equals(
        actual=output,
        expected=textwrap.dedent(
            """\
            usage: issue64 [-h] [--foo str] [--bar str]

            optional arguments:
                -h, --help  show this help message and exit

            Options ['options']:
                These are the options

                --foo str   Description (default: aaa)
                --bar str   (default: bbb)
            """
        ),
    )



import enum
from dataclasses import dataclass
from typing import Any, List, NamedTuple, Optional

import pytest
from typing_extensions import Literal

from .testutils import (
    TestSetup,
    raises_invalid_choice,
    raises_missing_required_arg,
    xfail_param,
)


class FieldComponents(NamedTuple):
    field_annotation: Any
    passed_value: Any
    parsed_value: Any
    incorrect_value: Any


class Color(enum.Enum):
    RED = enum.auto()
    BLUE = enum.auto()
    GREEN = enum.auto()


Fingers = Literal[0, 1, 2, 3, 4, 5]


@pytest.fixture(
    params=[
        FieldComponents(Literal["bob", "alice"], "bob", "bob", "clarice"),
        FieldComponents(Literal[True, False], "True", True, "bob"),
        xfail_param(
            [FieldComponents(Literal[True, False], "true", True, "bob")],
            reason="The support for boolean literals currently assumes just 'True' and 'False'.",
        ),
        FieldComponents(Literal[1, 2, 3], "1", 1, "foobar"),
        FieldComponents(Literal[1, 2, 3], "2", 2, "9"),
        FieldComponents(Literal["bob", "alice"], "bob", "bob", "clarice"),
        FieldComponents(Literal[Color.BLUE, Color.GREEN], "BLUE", Color.BLUE, "red"),
        FieldComponents(Literal[Color.BLUE, Color.GREEN], "BLUE", Color.BLUE, "foobar"),
        FieldComponents(Fingers, "1", 1, "foobar"),
        FieldComponents("Fingers", "1", 1, "foobar"),
    ]
)
def literal_field(request: pytest.FixtureRequest):
    field = request.param  # type: ignore
    return field


def test_literal(literal_field: FieldComponents):
    field_annotation, passed_value, parsed_value, incorrect_value = literal_field

    @dataclass
    class Foo(TestSetup):
        bar: field_annotation  # type: ignore

    with raises_missing_required_arg():
        Foo.setup("")

    assert Foo.setup(f"--bar {passed_value}") == Foo(bar=parsed_value)

    with raises_invalid_choice():
        assert Foo.setup(f"--bar {incorrect_value}")


@pytest.mark.xfail(reason="TODO: add support for optional literals")
def test_optional_literal(literal_field: FieldComponents):
    field_annotation, passed_value, parsed_value, incorrect_value = literal_field

    @dataclass
    class Foo(TestSetup):
        bar: Optional[field_annotation] = None  # type: ignore

    assert Foo.setup("") == Foo(bar=None)
    assert Foo.setup(f"--bar {passed_value}") == Foo(bar=parsed_value)

    with raises_invalid_choice():
        assert Foo.setup(f"--bar {incorrect_value}")


@pytest.mark.xfail(reason="TODO: Support lists of literals.")
def test_list_of_literal(literal_field: FieldComponents):
    field_annotation, passed_value, parsed_value, incorrect_value = literal_field

    @dataclass
    class Foo(TestSetup):
        values: List[field_annotation]  # type: ignore

    with raises_missing_required_arg():
        Foo.setup("")

    assert Foo.setup(f"--values {passed_value} {passed_value}") == Foo(
        values=[parsed_value, parsed_value]
    )
    with raises_invalid_choice():
        assert Foo.setup(f"--values {incorrect_value}")
